# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Battery project
"""

# %%
"""
## Day 2 - Li intercalation energy
"""

# %%
"""
Today we will calculate the energy cost/gain associated with intercalating a lithium atom into graphite using approaches at different levels of theory. After today you should be able to:

- Setup structures and do relaxations using ASE and GPAW.
- Discuss which level of theory is required to predict the Li intercalation energy and why?

The Li intercalation reaction is:
$$Li(s) + C_x^{graphite} \rightarrow LiC_x$$
and the intercalation energy will then be
$$E_{Li@graphite} = E_{LiC_x} - (E_{Li(s)} + E_{C_x^{graphite}})$$
Where $x$ is the number of Carbon atoms in your computational cell.

We will calculate these energies using Density Functional Theory (DFT) with different exchange-correlation functionals. Graphite is characterised by graphene layers with strongly bound carbon atoms in hexagonal rings and weak long-range van der Waals interactions between the layers.

In the sections below we will calculate the lowest energy structures of each compound. This is needed to determine the intercalation energy.

- [Graphite](#graphite)
- [Li metal](#limetal)
- [Li intercalation](#liintercalation)


"""

# %%
"""
<a id='graphite'></a>
## Graphite
"""

# %%
"""
![graphite](C64.png)
"""

# %%
"""
To try some of the methods that we are going to use we will start out by utilizing the interatomic potential [EMT](https://wiki.fysik.dtu.dk/ase/ase/calculators/emt.html) (Effective Medium Theory) calculator. This will allow to quickly test some of the optimization procedures in ASE, and ensure that the scripts do what they are supposed to, before we move on to the more precise and time consuming DFT calculations. Initially we will calculate the C-C distance and interlayer spacing in graphite in a two step procedure.
"""

# %%
# The graphite structure is set up using a tool in ASE
from ase.lattice.hexagonal import Graphite

# One has only to provide the lattice constants
structure = Graphite('C', latticeconstant={'a': 1.5, 'c': 4.0})

# %%
"""
To verify that the structures is as expected we can check it visually.
"""

# %%
from ase.visualize import view

view(structure)  # This will pop up the ase gui window

# %%
"""
Next we will use the EMT calculator to get the energy of graphite. Remember absolute energies are not meaningful, we will only use energy differences.
"""

# %%
from ase.calculators.emt import EMT

calc = EMT()
structure.calc = calc
energy = structure.get_potential_energy()
print(f'Energy of graphite: {energy:.2f} eV')


# %%
"""
The cell below requires some input. We set up a loop that should calculate the energy of graphite for a series of C-C distances.
"""

# %%
import numpy as np

# Provide some initial guesses on the distances
ccdist = 1.3
layerdist = 3.7

# We will explore distances around the initial guess
dists = np.linspace(ccdist * .8, ccdist * 1.2, 10)
# The resulting energies are put in the following list
energies = []
for cc in dists:
    # Calculate the lattice parameters a and c from the C-C distance
    # and interlayer distance, respectively
    a = cc * np.sqrt(3)
    c = 2 * layerdist

    gra = Graphite('C', latticeconstant={'a': a, 'c': c})

    # Set the calculator and attach it to the Atoms object with
    # the graphite structure
    calc = EMT()
    gra.calc = calc

    energy = gra.get_potential_energy()
    # Append results to the list
    energies.append(energy)


# %%
"""
Determine the equilibrium lattice constant. We use a [np.polynomial.Polynomial](https://www.numpy.org/devdocs/reference/generated/numpy.polynomial.polynomial.Polynomial.html#numpy.polynomial.polynomial.Polynomial)
"""

# %%
# Fit a polynomial:
poly = np.polynomial.Polynomial.fit(dists, energies, 3)
# and plot it:
# magic: %matplotlib notebook
import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 1)
ax.plot(dists, energies, '*r')
x = np.linspace(1, 2, 100)
ax.plot(x, poly(x))
ax.set_xlabel('C-C distance [Ang]')
ax.set_ylabel('energy [eV]')

# %%
poly1 = poly.deriv()
poly1.roots()  # two extrema

# %%
# Find the minimum:
emin, ccdist = min((poly(d), d) for d in poly1.roots())
ccdist

# %%
# alternatively:
poly2 = poly1.deriv()
for ccdist in poly1.roots():
    if poly2(ccdist) > 0:
        break
ccdist

# %%
"""
Make a script that calculates the interlayer distance with EMT in the cell below
"""

# %%
# This script will calculate the energy of graphite for a series of inter-layer distances.
from ase.calculators.emt import EMT
from ase.lattice.hexagonal import Graphite
from ase.eos import EquationOfState
import numpy as np

# ccdist is already defined in the previous cell
# Start from a small guess
layerdist = 2.0

# Teacher
dists = np.linspace(layerdist * .8, layerdist * 1.2, 10)
# The resulting energies are put in the following list
energies = []
for ld in dists:
    # Calculate the lattice parameters a and c from the C-C distance
    # and interlayer distance, respectively
    a = ccdist * np.sqrt(3)
    c = 2 * ld

    gra = Graphite('C', latticeconstant={'a': a, 'c': c})

    # Set the calculator and attach it to the Atoms object with
    # the graphite structure
    calc = EMT()
    gra.calc = calc

    energy = gra.get_potential_energy()
    # Append results to the list
    energies.append(energy)

poly = np.polynomial.Polynomial.fit(dists, energies, 3)
layerdist = next (d for d in poly.deriv().roots() if poly.deriv(2)(d) > 0)
print(layerdist)


# %%
"""
Now both the optimal C-C distance and the interlayer distance is evaluated with EMT. Try to compare with the experimental numbers provided below.

|                         | Experimental values | EMT | LDA | PBE | PBE+DFTD3 |
|-------------------------|---------------------|-----|-----|-----|-----------|
| C-C  distance / Å       |               1.420 |     |     |     |           |
| Interlayer distance / Å |                3.35 |     |     |     |           |

"""

# %%
"""
Not surprisingly we need to use more sophisticated theory to model this issue. Below we will use DFT as implemented in GPAW to determine the same parameters.
"""

# %%
"""
First we set up an initial guess of the structure as before.
"""

# %%
import numpy as np
from ase.lattice.hexagonal import Graphite

ccdist = 1.40
layerdist = 3.33
a = ccdist * np.sqrt(3)
c = 2 * layerdist
gra = Graphite('C', latticeconstant={'a': a, 'c': c})


# %%
"""
Then we create the GPAW calculator object. The parameters are explained [here](https://wiki.fysik.dtu.dk/gpaw/documentation/basic.html#parameters), see especially the section regarding [mode](https://wiki.fysik.dtu.dk/gpaw/documentation/basic.html#manual-mode). This graphite structure has a small unit cell, thus plane wave mode, `mode=PW()`, will be faster than the grid mode, `mode='fd'`. Plane wave mode also lets us calculate the strain on the unit cell - useful for optimizing the lattice parameters.

We will start by using the LDA exchange-correlation functional. Later you will try other functionals.
"""

# %%
from gpaw import GPAW, PW

xc = 'LDA'
calcname = f'graphite-{xc}'
calc = GPAW(mode=PW(500), kpts=(10, 10, 6), xc=xc,
            txt=calcname + '.log')

gra.calc = calc  # Connect system and calculator
print(gra.get_potential_energy())


# %%
"""
Check out the contents of the output file, all relevant information about the scf cycle are printed therein:
"""

# %%
n = calcname + '.log'
# magic: !cat $n

# %%
"""
Then we optimize the unit cell of the structure. We will take advantage of the [StrainFilter](https://wiki.fysik.dtu.dk/ase/ase/constraints.html#the-strainfilter-class) class. This allows us to simultaneously optimize both C-C distance and interlayer distance. We employ the [BFGS](http://aria42.com/blog/2014/12/understanding-lbfgs) algorithm to minimize the strain on the unit cell.
"""

# %%
from ase.constraints import StrainFilter
from ase.optimize.bfgs import BFGS
from ase.io import Trajectory

sf = StrainFilter(gra, mask=[1, 1, 1, 0, 0, 0])
opt = BFGS(sf, trajectory=calcname + '.traj')
# traj = Trajectory(calcname + '.traj', 'w', gra)
# opt.attach(traj)
opt.run(fmax=0.01)


# %%
"""
Read in the result of the relaxation and determine the C-C and interlayer distances.
"""

# %%
from ase.io import read
import numpy as np

atoms = read(calcname + '.traj')
a = atoms.cell[0, 0]
h = atoms.cell[2, 2]
# Determine the rest from here
print(a / np.sqrt(3))
print(h / 2)

# %%
"""
Now we need to try a GGA type functional (e.g. PBE) and also try to add van der Waals forces on top of PBE (i.e. PBE+DFTD3). These functionals will require more computational time, thus the following might be beneficial to read.
"""

# %%
"""
If the relaxation takes too long time, we can submit it to be run in parallel on the computer cluster. Remember we can then run different calculations simultaneously. There are two ways to achieve this:

A. Log in to the gbar terminal, save a full script in a file (e.g. `calc.py`) and submit that file to be calculated in parallel (e.g. `mq submit calc.py -R 8:5h` (5 hours on 8 cores)).

or

B. Stay in the notebook and submit the calculations using some commands we will give you, these are explained in the following cells.
"""

# %%
"""
You need to make a full working script in a cell that does not depend on variables defined in any other cells. Do that in the cell below. Once you are done, run it for a few seconds to make sure there is no error, then stop it by `interrupt the kernel`. This will give you an error message, ignore that and move on, we will submit the calculation in the next cell.
"""

# %%
# %%writefile graphite_LDA.py

# Full script
# from ase...

# Teacher
import numpy as np
from ase.lattice.hexagonal import Graphite
from ase.calculators.dftd3 import DFTD3
from ase.constraints import StrainFilter
from ase.optimize.bfgs import BFGS
from ase.io import Trajectory

ccdist = 1.41
layerdist = 3.21
a = ccdist * np.sqrt(3)
c = 2 * layerdist
gra = Graphite('C', latticeconstant={'a': a, 'c': c})

from gpaw import GPAW, PW

for xc in ['LDA', 'PBE', 'DFTD3']:
    if xc == 'DFTD3':
        dft = GPAW(mode=PW(500), kpts=(10, 10, 6), xc='PBE',
                   txt=f'graphite-{xc}.log')
        calc = DFTD3(dft=dft, xc='PBE')
    else:
        calc = GPAW(mode=PW(500), kpts=(10, 10, 6), xc=xc,
                    txt=f'graphite-{xc}.log')

    gra.calc = calc  # Connect system and calculator

    sf = StrainFilter(gra, mask=[1, 1, 1, 0, 0, 0])
    opt = BFGS(sf, trajectory=f'graphite-{xc}.traj')
    # traj = Trajectory(f'graphite-{xc}.traj', 'w', gra)
    # opt.attach(traj)
    opt.run(fmax=0.01)


# %%
"""
The following cell submits a calculation using [`myqueue`](https://myqueue.readthedocs.io/en/latest/).

Note the line `#%%writefile graphite_LDA.py` in the previous cell.  Remove the `#` symbol and execute the cell again.  This will write the cell to a file that you can use submit to the queue:
"""

# %%
# magic: !mq submit graphite_LDA.py -R 8:1h  # submits the calculation to 8 cores, 1 hour


# %%
"""
Check the status of the calculation by running the cell below. The column with `S` heading gives you a character code of the status: `Q`: still in queue, `R`: running, `C`: finished.
"""

# %%
# magic: !mq ls


# %%
"""
When the calculation finishes the result can be interpreted by [reading](https://wiki.fysik.dtu.dk/ase/ase/io/io.html#ase.io.read) in the Atoms object. This assumes that the trajectory file is called: `graphite-LDA.traj`, if not change accordingly.
"""

# %%
from ase.io import read

# The full relaxation
relaxation = read('graphite-LDA.traj', index=':')
# The final image of the relaxation containing the relevant energy
atoms = read('graphite-LDA.traj')

# %%
# Extract the relevant information from the calculation

# Energy
print(atoms.get_potential_energy())

# Unit cell
print(atoms.get_cell())

# See the steps of the optimization
from ase.visualize import view
view(relaxation)

# %%
"""
<a id='limetal'></a>
## Li metal
"""

# %%
"""
![Li metal](Li2.png)

Now we need to calculate the energy of Li metal. We will use the same strategy as for graphite, i.e. first determine the lattice constant, then use the energy of that structure. This time though you will have to do most of the work.

Some hints:
1. The crystal structure of Li metal is shown in the image above. That structure is easily created with one of the functions in the [ase.build](https://wiki.fysik.dtu.dk/ase/ase/build/build.html) module
2. A k-point density of approximately 2.4 points / Angstrom will be sufficient
3. The DFTD3 correction is done _a posteriori_, this means the calculator should be created a little differently, see [the second example here](https://wiki.fysik.dtu.dk/ase/ase/calculators/dftd3.html#examples)
4. See also the [equation of state module](https://wiki.fysik.dtu.dk/ase/ase/eos.html)

In the end try to compare the different functionals with experimental values:

|       | Experimental values | LDA | PBE | PBE+DFTD3 |
|-------|---------------------|-----|-----|-----------|
| a / Å |                3.51 |     |     |           |


"""

# %%
# Teacher

from ase import Atoms
from gpaw import GPAW, FermiDirac, PW
from ase.optimize import QuasiNewton
from ase.build import bulk
from ase.calculators.dftd3 import DFTD3
from ase.constraints import StrainFilter
from ase.optimize.bfgs import BFGS

# This script will optimize lattice constant of metallic lithium
for xc in ['LDA', 'PBE', 'DFTD3']:
    Li_metal = bulk('Li', crystalstructure='bcc', a=3.3)

    if xc == 'DFTD3':
        dft = GPAW(mode=PW(500),
                   kpts=(8, 8, 8),
                   nbands=-10,
                   txt=f'Li-metal-{xc}.log',
                   xc='PBE')
        calc = DFTD3(dft=dft, xc='PBE')
    else:
        calc = GPAW(mode=PW(500),
                    kpts=(8, 8, 8),
                    nbands=-10,
                    txt=f'Li-metal-{xc}.log',
                    xc=xc)

    Li_metal.calc = calc

    sf = StrainFilter(Li_metal, mask=[1, 1, 1, 0, 0, 0])
    opt = BFGS(sf, trajectory=f'Li-metal-{xc}.traj')
    opt.run(fmax=0.01)



# %%
"""
Get the lattice information and compare with experimental values
"""

# %%
a = Li_metal.get_cell()[0][0]  # student: a =


# %%
"""
<a id='liintercalation'></a>
## Li intercalation in graphite
"""

# %%
"""
![Li intercalated in graphite](C144Li18.png)

Now we will calculate the intercalation of Li in graphite. For simplicity we will represent the graphite with only one layer. Also try and compare the C-C and interlayer distances to experimental values.

|                         | Experimental values | LDA | PBE | PBE+DFTD3 |
|-------------------------|---------------------|-----|-----|-----------|
| C-C distance / Å        |               1.441 |     |     |           |
| Interlayer distance / Å |               3.706 |     |     |           |

"""

# %%
# A little help to get you started with the combined structure
from ase.lattice.hexagonal import Graphene
from ase import Atom

# We will have to optimize these distances
ccdist = 1.40
layerdist = 3.7

a = ccdist * np.sqrt(3)
c = layerdist

# We will require a larger cell, to accomodate the Li
Li_gra = Graphene('C', size=(2, 2, 1), latticeconstant={'a': a, 'c': c})
# Append an Li atom on top of the graphene layer
Li_gra.append(Atom('Li', (a / 2, ccdist / 2, layerdist / 2)))


# %%
# Teacher

from gpaw import GPAW, PW
from ase import Atom
from ase.optimize.bfgs import BFGS
import numpy as np
from ase.lattice.hexagonal import Graphene
from ase.constraints import StrainFilter

for xc in ['LDA', 'PBE', 'DFTD3']:
    ccdist = 1.40
    layerdist = 3.7

    a = ccdist * np.sqrt(3)
    c = layerdist

    calcname = f'Li-C8-{xc}'
    # We will require a larger cell, to accomodate the Li
    Li_gra = Graphene('C', size=(2, 2, 1), latticeconstant={'a': a, 'c': c})
    Li_gra.append(Atom('Li', (a / 2, ccdist / 2, layerdist / 2)))

    if xc == 'DFTD3':
        dft = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='PBE', txt=calcname + '.log')
        calc = DFTD3(dft=dft, xc='PBE')
    else:
        calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc=xc, txt=calcname + '.log')

    Li_gra.calc = calc  # Connect system and calculator

    sf = StrainFilter(Li_gra, mask=[1, 1, 1, 0, 0, 0])
    opt = BFGS(sf, trajectory=calcname + '.traj')
    opt.run(fmax=0.01)


# %%
"""
Now calculate the intercalation energy of Li in graphite with the following formula:
$$E_{Li@graphite} = E_{LiC_x} - (E_{Li(s)} + x * E_{C^{graphite}})$$
where $x$ is the number of Carbon atoms in your Li intercalated graphene cell. Remember to adjust the energies so that the correct number of atoms is taken into account.

These are the experimental values to compare with
In the end try to compare the different functionals with experimental values:

|                           | Experimental values | LDA | PBE | BEEF-vdW |
|---------------------------|---------------------|-----|-----|----------|
| Intercalation energy / eV |              -0.124 |     |     |          |


"""

# %%
# Teacher
e_Li_gra = Li_gra.get_potential_energy()
e_Li = Li_metal.get_potential_energy() / len(Li_metal)
e_C8 = 8 * gra.get_potential_energy() / len(gra)
intercalation_energy = e_Li_gra - (e_Li + e_C8)
print(f'Intercalation energy: {intercalation_energy:.2f}eV')

# %%
"""
Great job! When you have made it this far it is time to turn your attention to the cathode. If you have lots of time left though see if you can handle the bonus exercise.
"""

# %%
"""
## Bonus
"""

# %%
"""
In the calculation of the intercalated Li we used a graphene layer with 8 Carbon atoms per unit cell. We can actually use only 6 Carbon by rotating the x and y cell vectors. This structure will be faster to calculate and still have neglible Li-Li interaction.
"""

# %%
from ase import Atoms
from ase.visualize import view

ccdist = 1.40
layerdist = 3.7

a = ccdist * np.sqrt(3)
c = layerdist

Li_gra = Atoms('CCCCCCLi')  # Fill out the positions and cell vectors

# Teacher
for xc in ['LDA', 'PBE', 'DFTD3']:
    calcname = f'Li-C6-{xc}'
    Li_gra = Atoms('CCCCCCLi', positions=[(0, 0, 0), (0, ccdist, 0), (a, 0, 0),
                                          (-a, 0, 0), (-a / 2, -ccdist / 2, 0),
                                          (a / 2, -ccdist / 2, 0), (0, -ccdist, c / 2)],
                   cell=([1.5 * a, -1.5 * ccdist, 0],
                         [1.5 * a, 1.5 * ccdist, 0],
                         [0, 0, c]),
                   pbc=(1, 1, 1))

    if xc == 'DFTD3':
        dft = GPAW(mode=PW(500), kpts=(5, 5, 6), xc='PBE', txt=calcname + '.log')
        calc = DFTD3(dft=dft, xc='PBE')
    else:
        calc = GPAW(mode=PW(500), kpts=(5, 5, 6), xc=xc, txt=calcname + '.log')

    Li_gra.calc = calc  # Connect system and calculator

    sf = StrainFilter(Li_gra, mask=[1, 1, 1, 0, 0, 0])
    # traj = Trajectory(calcname + '.traj', 'w', Li_gra)
    opt = BFGS(sf, trajectory=calcname + '.traj')
    # opt.attach(traj)
    opt.run(fmax=0.01)

# %%



from ase import Atoms

from gpaw import GPAW, FermiDirac

k = 7
kT = 0
h = 0.18

a = 5.4
b = a / 2
atoms = Atoms('Si2',
              positions=([0, 0, 0],
                         [b / 2, b / 2, b / 2]),
              cell=([0, b, b],
                    [b, 0, b],
                    [b, b, 0]),
              pbc=True)

calc = GPAW(kpts=(k, k, k),
            setups='ah',
            symmetry='off',
            occupations=FermiDirac(kT),
            h=h)

atoms.calc = calc
atoms.get_potential_energy()
calc.write('Si_AH.gpw', mode='all')



import numpy as np
from gpaw.response.math_func import (two_phi_planewave_integrals,
                                     two_phi_nabla_planewave_integrals)
from gpaw.pw.lfc import PWLFC


def calculate_paw_corrections(setups, pd, spos_ac, soft):
    q_v = pd.K_qv[0]
    optical_limit = np.allclose(q_v, 0)

    G_Gv = pd.get_reciprocal_vectors()
    if optical_limit:
        G_Gv[0] = 1

    pos_av = spos_ac @ pd.gd.cell_cv

    # Collect integrals for all species:
    Q_xGii = {}
    for id, atomdata in setups.setups.items():
        if soft:
            ghat = PWLFC([atomdata.ghat_l], pd)
            ghat.set_positions(np.zeros((1, 3)))
            Q_LG = ghat.expand().T
            if atomdata.Delta_iiL is None:
                ni = atomdata.ni
                Q_Gii = np.zeros((Q_LG.shape[1], ni, ni))
            else:
                Q_Gii = np.dot(atomdata.Delta_iiL, Q_LG).T
        else:
            ni = atomdata.ni
            Q_Gii = two_phi_planewave_integrals(G_Gv, atomdata)
            Q_Gii.shape = (-1, ni, ni)

        Q_xGii[id] = Q_Gii

    Q_aGii = []
    for a, atomdata in enumerate(setups):
        id = setups.id_a[a]
        Q_Gii = Q_xGii[id]
        x_G = np.exp(-1j * np.dot(G_Gv, pos_av[a]))
        Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Q_Gii)
        if optical_limit:
            Q_aGii[a][0] = atomdata.dO_ii

    return Q_aGii


def calculate_paw_nabla_corrections(setups, spos_ac, pd, soft):
    G_Gv = pd.get_reciprocal_vectors()
    pos_av = np.dot(spos_ac, pd.gd.cell_cv)

    # Collect integrals for all species:
    Q_xvGii = {}
    for id, atomdata in setups.setups.items():
        if soft:
            raise NotImplementedError
        else:
            Q_vGii = two_phi_nabla_planewave_integrals(G_Gv, atomdata)
            ni = atomdata.ni
            Q_vGii.shape = (3, -1, ni, ni)

        Q_xvGii[id] = Q_vGii

    Q_avGii = []
    for a, atomdata in enumerate(setups):
        id = setups.id_a[a]
        Q_vGii = Q_xvGii[id]
        x_G = np.exp(-1j * np.dot(G_Gv, pos_av[a]))
        Q_avGii.append(x_G[np.newaxis, :, np.newaxis, np.newaxis] * Q_vGii)

    return Q_avGii



from gpaw.new.ase_interface import GPAW
from ase import Atoms
import numpy as np

a = 6.339
d = 1.331

atoms = Atoms('V3Cl6',
              cell=[a, a, 1, 90, 90, 60],
              pbc=[1, 1, 0],
              scaled_positions=[
                  [0, 0, 0],
                  [1 / 3, 1 / 3, 0],
                  [2 / 3, 2 / 3, 0],
                  [0, 2 / 3, d],
                  [0, 1 / 3, -d],
                  [1 / 3, 0, d],
                  [1 / 3, 2 / 3, -d],
                  [2 / 3, 1 / 3, d],
                  [2 / 3, 0, -d]])
atoms.center(axis=2, vacuum=5)

m = 3.0
magmoms = np.zeros((9, 3))
magmoms[0] = [m, 0, 0]
magmoms[1] = [-m / 2, m * 3**0.5 / 2, 0]
magmoms[2] = [-m / 2, -m * 3**0.5 / 2, 0]

atoms.calc = GPAW(mode={'name': 'pw',
                        'ecut': 400},
                  magmoms=magmoms,
                  symmetry='off',
                  kpts={'size': (2, 2, 1), 'gamma': True},
                  parallel={'domain': 1, 'band': 1},
                  txt='VCl2_gs.txt')

atoms.get_potential_energy()
atoms.calc.write('VCl2_gs.gpw')



def workflow():
    from myqueue.workflow import run
    with run(script='Al_fcc.py', cores=2, tmax='15m'):
        run(script='../../electronic/dos/dos.py',
            args=['Al-fcc.gpw'])
    run(script='Al_bcc.py', cores=2, tmax='15m')
    run(script='Al_fcc_vs_bcc.py', cores=2)
    run(script='Al_fcc_modified.py', cores=2)



import os
import gpaw.wannier90 as w90
from gpaw import GPAW

seed = 'Cu'

calc = GPAW(seed + '.gpw', txt=None)

w90.write_input(calc,
                orbitals_ai=[[0, 1, 4, 5, 6, 7, 8]],
                bands=range(20),
                num_iter=1000,
                dis_num_iter=500)

os.system('wannier90.x -pp ' + seed)

w90.write_projections(calc, orbitals_ai=[[0, 1, 4, 5, 6, 7, 8]])
w90.write_eigenvalues(calc)
w90.write_overlaps(calc)

os.system('wannier90.x ' + seed)



import numpy as np
from ase import Atoms
from ase.units import Bohr
from gpaw.jellium import Jellium
from gpaw import GPAW, PW

rs = 5.0 * Bohr  # Wigner-Seitz radius
h = 0.2          # grid-spacing
a = 8 * h        # lattice constant
k = 12           # number of k-points (k*k*k)

ne = a**3 / (4 * np.pi / 3 * rs**3)
jellium = Jellium(ne)

bulk = Atoms(pbc=True, cell=(a, a, a))
bulk.calc = GPAW(mode=PW(400.0),
                 background_charge=jellium,
                 xc='LDA_X+LDA_C_WIGNER',
                 nbands=5,
                 kpts=[k, k, k],
                 h=h,
                 txt='bulk.txt')
e0 = bulk.get_potential_energy()



# web-page: dipoles.png
import numpy as np
import matplotlib.pyplot as plt

fig, ax = plt.subplots(figsize=(6, 6 / 1.62))

data = np.loadtxt('dm_nad.dat')
ax.plot(data[:, 0], data[:, 4], label=r'Na$_2$')
data = np.loadtxt('dm_nad2.dat')
ax.plot(data[:, 0], data[:, 4], label=r'2Na$_2$')
plt.xlabel('Time')
plt.legend()
plt.ylabel(r'<z>')
plt.tight_layout()
plt.savefig('dipoles.png')



from ase.build import bulk
from gpaw import GPAW, PW

cu = bulk('Cu', 'fcc', a=3.6)

for smearing in [{'name': 'improved-tetrahedron-method'},
                 {'name': 'tetrahedron-method'},
                 {'name': 'fermi-dirac', 'width': 0.05},
                 {'name': 'marzari-vanderbilt', 'width': 0.2}]:
    name = ''.join(word[0].upper() for word in smearing['name'].split('-'))
    width = smearing.get('width')
    if width:
        name += f'-{width}'

    for k in range(8, 21):
        cu.calc = GPAW(
            mode=PW(400),
            kpts=(k, k, k),
            occupations=smearing,
            txt=f'Cu-{name}-{k}.txt')
        e = cu.get_potential_energy()



from ase.parallel import paropen
from gpaw.xc.rpa import RPACorrelation

ds = [1.75, 2.0, 2.25, 2.5, 2.75, 3.0, 3.25, 3.5, 3.75, 4.0, 5.0, 6.0, 10.0]
ecut = 200

for d in ds:
    rpa = RPACorrelation(f'gs_{d}.gpw', txt=f'rpa_{ecut}_{d}.txt')
    E_rpa = rpa.calculate(ecut=[ecut],
                          frequency_scale=2.5,
                          skip_gamma=True,
                          filename=f'restart_{ecut}_{d}.txt')

    f = paropen(f'rpa_{ecut}.dat', 'a')
    print(d, E_rpa, file=f)
    f.close()



import numpy as np
from ase.build import bulk
from gpaw import GPAW, PW

a0 = 4.04
al = bulk('Al', 'fcc', a=a0)
cell0 = al.cell

for ecut in range(200, 501, 50):
    al.calc = GPAW(mode=PW(ecut),
                   xc='PBE',
                   kpts=(8, 8, 8),
                   basis='dzp',
                   txt=f'Al-{ecut}.txt')
    for eps in np.linspace(-0.02, 0.02, 5):
        al.cell = (1 + eps) * cell0
        al.get_potential_energy()

al.calc.set(mode=PW(400))
for k in range(4, 17):
    al.calc.set(kpts=(k, k, k),
                txt=f'Al-{k:02}.txt')
    for eps in np.linspace(-0.02, 0.02, 5):
        al.cell = (1 + eps) * cell0
        al.get_potential_energy()



from myqueue.workflow import run


def workflow():
    with run(script='gs.py', cores=8):
        with run(script='unocc.py', cores=8, tmax='30m'):
            with run(script='lr2.py', cores=8):
                with run(script='lr2_restart.py', cores=8):
                    with run(script='lr2_analyze.py', cores=1):
                        run(script='plot_spec.py')
                        run(script='check_consistency_with_docs.py')



import math

import numpy as np

from ase import Atom, Atoms
from ase.build import molecule

systems = {}

name = 'dichloroallene'
system = Atoms()
CC = 1.33  # C-C bond length
CH = 1.1  # C-H bond length
CCl = 1.74  # C-Cl bond length
system.append(Atom('C', position=[0, 0, 0]))
system.append(Atom('C', position=[CC, 0, 0]))
system.append(Atom('C', position=[-CC, 0, 0]))
c = 1 / math.sqrt(3.0)
system.append(Atom('Cl', position=[CC + c * CCl, c * CCl, c * CCl]))
system.append(Atom('H', position=[CC + c * CH, -c * CH, -c * CH]))
system.append(Atom('Cl', position=[-CC - c * CCl, -c * CCl, c * CCl]))
system.append(Atom('H', position=[-CC - c * CH, +c * CH, -c * CH]))
systems['C2'] = (name, system)

name = 'H2O'
a = 8.0
h = 0.2
system = molecule(name)
systems['C2v'] = (name, system)

name = 'NH3'
deg2rad = 2 * math.pi / 360.
NH = 1.008  # N-H bond length
alpha = 107.5 * deg2rad  # H-N-H angle
system = Atoms()
system.append(Atom('N', position=[0, 0, 0]))
x = NH * math.sin(alpha / 2.) / math.cos(math.pi / 6.)
y = 0.
z = math.sqrt(NH**2 - x**2)
atom0 = Atoms()
atom0.append(Atom('H', position=[x, y, z]))
beta = 2 * math.pi / 3.
rotmat = np.array([[math.cos(beta), -math.sin(beta), 0.],
                   [+math.sin(beta), math.cos(beta), 0.], [0., 0., 1.]])
for i in range(3):
    atom0[0].position = rotmat.dot(atom0[0].position)
    system.append(atom0[0])
systems['C3v'] = (name, system)

name = 'allene'
system = Atoms()
CC = 1.33  # C-C bond length
CH = 1.1  # C-H bond length
system.append(Atom('C', position=[0, 0, 0]))
system.append(Atom('C', position=[0, 0, CC]))
system.append(Atom('C', position=[0, 0, -CC]))
c = 1 / np.sqrt(2.0)
system.append(Atom('H', position=[0., c * CH, CC + c * CH]))
system.append(Atom('H', position=[0., -c * CH, CC + c * CH]))
system.append(Atom('H', position=[c * CH, 0., -CC - c * CH]))
system.append(Atom('H', position=[-c * CH, 0., -CC - c * CH]))
systems['D2d'] = (name, system)

name = 'BF3'
deg2rad = 2 * math.pi / 360.
BF = 1.313  # B-F bond length
alpha = 120. * deg2rad  # F-B-F angle
system = Atoms()
system.append(Atom('B', position=[0, 0, 0]))
x = BF * math.sin(alpha / 2.) / math.cos(math.pi / 6.)
y = 0.
z = math.sqrt(BF**2 - x**2)
atom0 = Atoms()
atom0.append(Atom('F', position=[x, y, z]))
beta = 2 * math.pi / 3.
rotmat = np.array([[math.cos(beta), -math.sin(beta), 0.],
                   [+math.sin(beta), math.cos(beta), 0.], [0., 0., 1.]])
for i in range(3):
    atom0[0].position = rotmat.dot(atom0[0].position)
    system.append(atom0[0])
systems['D3h'] = (name, system)

name = 'ferrocene'
system = Atoms()
system.append(Atom('Fe', position=[0, 0, 0]))
height = 1.68  # distance of cyclopentadienyl from Fe
CC = 1.425  # C-C distance
CH = 1.080  # C-H distance
FeC_xy = CC / (2. * math.sin(2 * math.pi / 5. / 2.)
               )  # distance of C from z axis
chirality_angle = 0. * (2 * math.pi / 360.)
# Top ring:
for i in range(5):
    x = FeC_xy * math.cos(i * 2 * math.pi / 5. + chirality_angle / 2.)
    y = FeC_xy * math.sin(i * 2 * math.pi / 5. + chirality_angle / 2.)
    z = height
    system.append(Atom('C', position=[x, y, z]))

    x = (FeC_xy + CH) * math.cos(i * 2 * math.pi / 5. + chirality_angle / 2.)
    y = (FeC_xy + CH) * math.sin(i * 2 * math.pi / 5. + chirality_angle / 2.)
    z = height
    system.append(Atom('H', position=[x, y, z]))
# Bottom ring:
for i in range(5):
    x = FeC_xy * math.cos(i * 2 * math.pi / 5. - chirality_angle / 2.)
    y = FeC_xy * math.sin(i * 2 * math.pi / 5. - chirality_angle / 2.)
    z = -height
    system.append(Atom('C', position=[x, y, z]))
    x = (FeC_xy + CH) * math.cos(i * 2 * math.pi / 5. - chirality_angle / 2.)
    y = (FeC_xy + CH) * math.sin(i * 2 * math.pi / 5. - chirality_angle / 2.)
    z = -height
    system.append(Atom('H', position=[x, y, z]))
systems['D5h'] = (name, system)

name = 'ferrocene-chiral'
system = Atoms()
system.append(Atom('Fe', position=[0, 0, 0]))
height = 1.68  # distance of cyclopentadienyl from Fe
CC = 1.425  # C-C distance
CH = 1.080  # C-H distance
FeC_xy = CC / (2. * math.sin(2 * math.pi / 5. / 2.)
               )  # distance of C from z axis
chirality_angle = 14. * (2 * math.pi / 360.)
# Top ring:
for i in range(5):
    x = FeC_xy * math.cos(i * 2 * math.pi / 5. + chirality_angle / 2.)
    y = FeC_xy * math.sin(i * 2 * math.pi / 5. + chirality_angle / 2.)
    z = height
    system.append(Atom('C', position=[x, y, z]))

    x = (FeC_xy + CH) * math.cos(i * 2 * math.pi / 5. + chirality_angle / 2.)
    y = (FeC_xy + CH) * math.sin(i * 2 * math.pi / 5. + chirality_angle / 2.)
    z = height
    system.append(Atom('H', position=[x, y, z]))
# Bottom ring:
for i in range(5):
    x = FeC_xy * math.cos(i * 2 * math.pi / 5. - chirality_angle / 2.)
    y = FeC_xy * math.sin(i * 2 * math.pi / 5. - chirality_angle / 2.)
    z = -height
    system.append(Atom('C', position=[x, y, z]))

    x = (FeC_xy + CH) * math.cos(i * 2 * math.pi / 5. - chirality_angle / 2.)
    y = (FeC_xy + CH) * math.sin(i * 2 * math.pi / 5. - chirality_angle / 2.)
    z = -height
    system.append(Atom('H', position=[x, y, z]))
systems['D5'] = (name, system)

name = 'dodecaborate'
# Icosahedron params:
BB = 1.8  # B-B bond length
BH = 1.2  # B-H bond length
phi = 0.5 * (1. + math.sqrt(5.))
l = 1.0  # streching parameter
P = BB / 2.  # edge length parameter
ico = np.array([[0., l, phi], [0., l, -phi], [0., -l, phi], [0., -l, -phi],
                [l, phi, 0.], [l, -phi, 0.], [-l, phi, 0.], [-l, -phi, 0.],
                [phi, 0., l], [-phi, 0., l], [phi, 0., -l], [-phi, 0., -l]
                ]) * P
system = Atoms()
for corner in ico:
    system.append(Atom('B', position=corner))
    Hpos = corner + BH * corner / np.linalg.norm(corner)
    system.append(Atom('H', position=Hpos))
R = system.positions
system.rotate(R[5] - R[3], 'z')
system.rotate(36, 'z')
systems['Ih'] = (name, system)

name = 'dodecaborate'
# Icosahedron params:
BB = 1.8  # B-B bond length
BH = 1.2  # B-H bond length
phi = 0.5 * (1. + math.sqrt(5.))
l = 1.0  # streching parameter
P = BB / 2.  # edge length parameter
ico = np.array([[0., l, phi], [0., l, -phi], [0., -l, phi], [0., -l, -phi],
                [l, phi, 0.], [l, -phi, 0.], [-l, phi, 0.], [-l, -phi, 0.],
                [phi, 0., l], [-phi, 0., l], [phi, 0., -l], [-phi, 0., -l]
                ]) * P
system = Atoms()
for corner in ico:
    system.append(Atom('B', position=corner))
    Hpos = corner + BH * corner / np.linalg.norm(corner)
    system.append(Atom('H', position=Hpos))
systems['Ico'] = (name, system)

name = 'F6S'
system = Atoms()
SF = 1.564  # S-F bond length
system.append(Atom('S', position=[0, 0, 0]))
system.append(Atom('F', position=[0., 0., SF]))
system.append(Atom('F', position=[0., 0., -SF]))
system.append(Atom('F', position=[0., SF, 0.]))
system.append(Atom('F', position=[0., -SF, 0.]))
system.append(Atom('F', position=[SF, 0., 0.]))
system.append(Atom('F', position=[-SF, 0., 0.]))
systems['Oh'] = (name, system)

name = 'P4'
PP = 2.234  # P-P bond length
alpha = 60. * deg2rad
system = Atoms()
system.append(Atom('P', position=[0, 0, 0]))
x = PP * math.sin(alpha / 2.) / math.cos(math.pi / 6.)
y = 0.
z = -math.sqrt(PP**2 - x**2)
atom0 = Atoms()
atom0.append(Atom('P', position=[x, y, z]))
beta = 2 * math.pi / 3.
rotmat = np.array([[math.cos(beta), -math.sin(beta), 0.],
                   [+math.sin(beta), math.cos(beta), 0.], [0., 0., 1.]])
for i in range(3):
    atom0[0].position = rotmat.dot(atom0[0].position)
    system.append(atom0[0])
system.rotate(180, 'z')
systems['Td'] = (name, system)

name = 'Mg-H2O-6'
MgO = 2.09  # Mg-O bond length
atoms = Atoms()
atoms.append(Atom('Mg', position=[0, 0, 0]))
h2o = molecule('H2O')
opos = h2o[0].position.copy()
for atom in h2o:
    atom.position -= opos
for atom in h2o:
    atom.position += np.array([0., 0., -MgO])
for atom in h2o:
    atoms.append(atom)
    atom.position *= -1
    atoms.append(atom)
for atom in h2o:
    atom.position = np.array(
        [atom.position[1], atom.position[0], atom.position[2]])
    atom.position = np.array(
        [atom.position[2], atom.position[1], atom.position[0]])
    atoms.append(atom)
    atom.position *= -1
    atoms.append(atom)
for atom in h2o:
    atom.position = np.array(
        [atom.position[0], atom.position[2], atom.position[1]])
    atom.position = np.array(
        [atom.position[1], atom.position[0], atom.position[2]])
    atoms.append(atom)
    atom.position *= -1
    atoms.append(atom)
system = atoms
system.rotate([1, 1, 1], 'z')
system.rotate(75, 'z')
systems['Th'] = (name, system)

for group_name, (name, system) in systems.items():
    system.write(f'{group_name}-{name}.xyz')



import os

parallel_python_interpreter=False

#compiler = 'gcc'
mpicompiler = 'mpicc'
scalapack = True
fftw = True

libraries = ['mkl_intel_lp64',
            'mkl_sequential',
            'mkl_lapack',
             'mkl_core',
             'pthread',
             'readline',
             'termcap',
             'xc',
             'mkl_blacs_intelmpi_lp64']

# FFTW should be configured from environment variables, but they do
# not report the correct names for a dynamically loaded library.
# Use Intel MKL
if fftw:
    libraries += ['fftw3xc_intel_pic', 'mkl_rt']

# Use EasyBuild scalapack from the active toolchain
if scalapack:
    libraries += ['mkl_scalapack_lp64','mkl_lapack95_lp64']

# Use EasyBuild libxc
libxc = os.getenv('EBROOTLIBXC')
include_dirs.append(os.path.join(libxc, 'include'))

# libvdwxc:
# Use EasyBuild libvdwxc
# NOTE: This currenlty does not work together with the Intel MKL, so
# the easyconfig files does not load libvdwxc
libvdwxc = os.getenv('EBROOTLIBVDWXC')
if libvdwxc:
    include_dirs.append(os.path.join(libvdwxc, 'include'))
    libraries.append('vdwxc')

# ELPA:
# Use EasyBuild ELPA if loaded
elpa = os.getenv('EBROOTELPA')
if elpa:
    libraries += ['elpa']
    elpaversion = os.path.basename(elpa).split('-')[0]
    library_dirs.append(os.path.join(elpa, 'lib'))
    extra_link_args += ['-Wl,-rpath={}/lib'.format(elpa)]
    include_dirs.append(os.path.join(elpa, 'include', 'elpa-'+elpaversion))

# Now add a EasyBuild "cover-all-bases" library_dirs
library_dirs = os.getenv('LD_LIBRARY_PATH').split(':')



import pytest
from ase.units import Bohr

from gpaw.calculator import GPAW as GPAW1
from gpaw.new.ase_interface import GPAW as GPAW2
from gpaw.utilities.ps2ae import PS2AE
from gpaw.mpi import size


@pytest.mark.parametrize('name, tol',
                         [('bcc_li_pw_wfs', 3e-5),
                          ('bcc_li_fd_wfs', 4e-4)])
def test_ae_k(gpw_files, name, tol):
    """Test normalization of non gamma-point wave functions."""

    # Old API:
    calc = GPAW1(gpw_files[name])
    converter = PS2AE(calc)
    ae = converter.get_wave_function(n=0, k=1, ae=True) * Bohr**1.5
    norm = converter.gd.integrate((ae * ae.conj()).real)
    assert norm == pytest.approx(1.0, abs=tol)

    # New API:
    if size > 1:
        return
    calc = GPAW2(gpw_files[name])
    ae = calc.calculation.state.ibzwfs.get_all_electron_wave_function(
        0, kpt=1, grid_spacing=0.1)
    assert ae.norm2() == pytest.approx(1.0, abs=tol)



# creates: nitrogen-log-derivs.png
from gpaw.atom.generator2 import generate, plot_log_derivs
import matplotlib.pyplot as plt

kwargs = {'symbol': 'N',
          'Z': 7,
          'xc': 'PBE',
          'projectors': '2s,s,2p,p,d,F',
          'radii': [1.3],
          'scalar_relativistic': True,
          'r0': 1.1,
          'v0': None,
          'nderiv0': 5,
          'pseudize': ('poly', 4)}

g = generate(**kwargs)
plot_log_derivs(g, 'spdfg', True)
plt.savefig('nitrogen-log-derivs.png')



extra_compile_args = ['-std=c99', '-O3']
compiler = 'cc'
mpicompiler = 'cc'
mpilinker= 'cc'
# edit library and include paths for libxc
include_dirs += ['/homeappl/home/jenkovaa/libxc/sisu/gcc/include']
library_dirs = ['/homeappl/home/jenkovaa/libxc/sisu/gcc/lib']


libraries = ['z', 'xc']

scalapack = True
hdf5 = True

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
define_macros += [("GPAW_ASYNC",1)]
define_macros += [("GPAW_MPI2",1)]



from myqueue.workflow import run


def workflow():
    with run(script='gs_BaTiO3.py', cores=8, tmax='30m'):
        run(script='polarization_BaTiO3.py', cores=8, tmax='1h')
        with run(script='born_BaTiO3.py', cores=8, tmax='10h'):
            run(script='get_borncharges.py')

    with run(script='gs_Sn.py', cores=8, tmax='30m'):
        with run(script='Sn_parallel_transport.py', cores=8, tmax='5h'):
            run(script='plot_phase.py')



import sys
from time import ctime
from pathlib import Path
import pickle

import numpy as np

from ase.units import Hartree
from gpaw.utilities import convert_string_to_fd
from ase.utils.timing import Timer, timer

import gpaw.mpi as mpi
from gpaw.response.kspair import get_calc
from gpaw.response.frequencies import FrequencyDescriptor
from gpaw.response.chiks import ChiKS
from gpaw.response.kxc import get_fxc
from gpaw.response.kernels import get_coulomb_kernel
from gpaw.response.pw_parallelization import Blocks1D
from gpaw.response.groundstate import ResponseGroundStateAdapter


class FourComponentSusceptibilityTensor:
    """Class calculating the full four-component susceptibility tensor"""

    def __init__(self, gs, fxc='ALDA', fxckwargs={},
                 eta=0.2, ecut=50, gammacentered=False,
                 disable_point_group=False, disable_time_reversal=False,
                 bandsummation='pairwise', nbands=None, bundle_integrals=True,
                 world=mpi.world, nblocks=1, txt=sys.stdout):
        """
        Currently, everything is in plane wave mode.
        If additional modes are implemented, maybe look to fxc to see how
        multiple modes can be supported.

        Parameters
        ----------
        gs : see gpaw.response.chiks, gpaw.response.kslrf
        fxc, fxckwargs : see gpaw.response.fxc
        eta, ecut, gammacentered
        disable_point_group,
        disable_time_reversal,
        bandsummation, nbands, bundle_integrals,
        world, nblocks, txt : see gpaw.response.chiks, gpaw.response.kslrf
        """
        # Initiate output file and timer
        self.world = world
        self.fd = convert_string_to_fd(txt, world)
        self.cfd = self.fd
        self.timer = Timer()

        # Load ground state calculation
        self.calc = get_calc(gs, fd=self.fd, timer=self.timer)
        self.gs = ResponseGroundStateAdapter(self.calc)

        # The plane wave basis is defined by keywords
        self.ecut = None if ecut is None else ecut / Hartree
        self.gammacentered = gammacentered

        # Initiate Kohn-Sham susceptibility and fxc objects
        self.chiks = ChiKS(self.calc, eta=eta, ecut=ecut,
                           gammacentered=gammacentered,
                           disable_point_group=disable_point_group,
                           disable_time_reversal=disable_time_reversal,
                           bandsummation=bandsummation, nbands=nbands,
                           bundle_integrals=bundle_integrals,
                           world=world, nblocks=nblocks, txt=self.fd,
                           timer=self.timer)
        self.fxc = get_fxc(self.calc, fxc,
                           response='susceptibility', mode='pw',
                           world=self.chiks.world, txt=self.chiks.fd,
                           timer=self.timer, **fxckwargs)

        # Parallelization over frequencies depends on the frequency input
        self.blocks1d = None

    def get_macroscopic_component(self, spincomponent, q_c, frequencies,
                                  filename=None, txt=None):
        """Calculates the spatially averaged (macroscopic) component of the
        susceptibility tensor and writes it to a file.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        filename : str
            Save chiks_w and chi_w to file of given name.
            Defaults to:
            'chi%s_q«%+d-%+d-%+d».csv' % (spincomponent,
                                          *tuple((q_c * kd.N_c).round()))
        txt : str
            Save output of the calculation of this specific component into
            a file with the filename of the given input.

        Returns
        -------
        see calculate_macroscopic_component
        """

        if filename is None:
            tup = (spincomponent,
                   *tuple((q_c * self.calc.wfs.kd.N_c).round()))
            filename = 'chi%s_q«%+d-%+d-%+d».csv' % tup

        (omega_w,
         chiks_w,
         chi_w) = self.calculate_macroscopic_component(spincomponent, q_c,
                                                       frequencies,
                                                       txt=txt)

        write_macroscopic_component(omega_w, chiks_w, chi_w,
                                    filename, self.world)

        return omega_w, chiks_w, chi_w

    def calculate_macroscopic_component(self, spincomponent,
                                        q_c, frequencies, txt=None):
        """Calculates the spatially averaged (macroscopic) component of the
        susceptibility tensor.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        txt : see get_macroscopic_component

        Returns
        -------
        omega_w, chiks_w, chi_w : nd.array, nd.array, nd.array
            omega_w: frequencies in eV
            chiks_w: macroscopic dynamic susceptibility (Kohn-Sham system)
            chi_w: macroscopic dynamic susceptibility
        """
        (pd, wd,
         chiks_wGG, chi_wGG) = self.calculate_component(spincomponent, q_c,
                                                        frequencies, txt=txt)

        # Macroscopic component
        chiks_w = chiks_wGG[:, 0, 0]
        chi_w = chi_wGG[:, 0, 0]

        # Collect data for all frequencies
        omega_w = wd.omega_w * Hartree
        chiks_w = self.collect(chiks_w)
        chi_w = self.collect(chi_w)

        return omega_w, chiks_w, chi_w

    def get_component_array(self, spincomponent, q_c, frequencies,
                            array_ecut=50, filename=None, txt=None):
        """Calculates a specific spin component of the susceptibility tensor,
        collects it as a numpy array in a reduced plane wave description
        and writes it to a file.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        array_ecut : see calculate_component_array
        filename : str
            Save chiks_w and chi_w to pickle file of given name.
            Defaults to:
            'chi%sGG_q«%+d-%+d-%+d».pckl' % (spincomponent,
                                             *tuple((q_c * kd.N_c).round()))
        txt : str
            Save output of the calculation of this specific component into
            a file with the filename of the given input.

        Returns
        -------
        see calculate_component_array
        """

        if filename is None:
            tup = (spincomponent,
                   *tuple((q_c * self.gs.kd.N_c).round()))
            filename = 'chi%sGG_q«%+d-%+d-%+d».pckl' % tup

        (omega_w, G_Gc, chiks_wGG,
         chi_wGG) = self.calculate_component_array(spincomponent,
                                                   q_c,
                                                   frequencies,
                                                   array_ecut=array_ecut,
                                                   txt=txt)

        # Write susceptibilities to a pickle file
        write_component(omega_w, G_Gc, chiks_wGG, chi_wGG,
                        filename, self.world)

        return omega_w, G_Gc, chiks_wGG, chi_wGG

    def calculate_component_array(self, spincomponent, q_c, frequencies,
                                  array_ecut=50, txt=None):
        """Calculates a specific spin component of the susceptibility tensor
        and collects it as a numpy array in a reduced plane wave description.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf
        array_ecut : float
            Energy cutoff for the reduced plane wave representation.
            The susceptibility is returned in the reduced representation.

        Returns
        -------
        omega_w, G_Gc, chiks_wGG, chi_wGG : nd.array(s)
            omega_w: frequencies in eV
            G_Gc : plane wave repr. as coordinates on the reciprocal lattice
            chiks_wGG: dynamic susceptibility (Kohn-Sham system)
            chi_wGG: dynamic susceptibility
        """
        (pd, wd,
         chiks_wGG, chi_wGG) = self.calculate_component(spincomponent, q_c,
                                                        frequencies, txt=txt)

        # Get frequencies in eV
        omega_w = wd.omega_w * Hartree

        # Get susceptibility in a reduced plane wave representation
        mask_G = get_pw_reduction_map(pd, array_ecut)
        chiks_wGG = np.ascontiguousarray(chiks_wGG[:, mask_G, :][:, :, mask_G])
        chi_wGG = np.ascontiguousarray(chi_wGG[:, mask_G, :][:, :, mask_G])

        # Get reduced plane wave repr. as coordinates on the reciprocal lattice
        G_Gc = get_pw_coordinates(pd)[mask_G]

        # Gather susceptibilities for all frequencies
        chiks_wGG = self.gather(chiks_wGG, wd)
        chi_wGG = self.gather(chi_wGG, wd)

        return omega_w, G_Gc, chiks_wGG, chi_wGG

    def calculate_component(self, spincomponent, q_c, frequencies, txt=None):
        """Calculate a single component of the susceptibility tensor.

        Parameters
        ----------
        spincomponent, q_c,
        frequencies : see gpaw.response.chiks, gpaw.response.kslrf

        Returns
        -------
        pd : PWDescriptor
            Descriptor object for the plane wave basis
        wd : FrequencyDescriptor
            Descriptor object for the calculated frequencies
        chiks_wGG : ndarray
            The process' block of the Kohn-Sham susceptibility component
        chi_wGG : ndarray
            The process' block of the full susceptibility component
        """

        # Initiate new call-output file, if supplied
        if txt is not None:
            # Store timer and close old call-output file
            self.write_timer()
            if str(self.fd) != str(self.cfd):
                print('\nClosing, %s' % ctime(), file=self.cfd)
                self.cfd.close()
            # Initiate new output file
            self.cfd = convert_string_to_fd(txt, self.world)
        # Print to output file(s)
        if str(self.fd) != str(self.cfd):
            print('---------------', file=self.fd)
            print(f'Calculating susceptibility spincomponent={spincomponent}'
                  f'with q_c={q_c}', flush=True, file=self.fd)
        if txt is not None:
            print('---------------', file=self.fd)
            print(f'Calculating susceptibility spincomponent={spincomponent}'
                  f'with q_c={q_c}', file=self.cfd)
            print('---------------', flush=True, file=self.cfd)

        wd = FrequencyDescriptor.from_array_or_dict(frequencies)

        # Initialize parallelization over frequencies
        self.blocks1d = Blocks1D(self.world, len(wd))

        return self._calculate_component(spincomponent, q_c, wd)

    def _calculate_component(self, spincomponent, q_c, wd):
        """In-place calculation of the given spin-component."""
        pd, chiks_wGG = self.calculate_ks_component(spincomponent, q_c,
                                                    wd, txt=self.cfd)

        Kxc_GG = self.get_xc_kernel(spincomponent, pd, chiks_wGG=chiks_wGG)
        if spincomponent in ['+-', '-+']:
            # No Hartree kernel
            assert Kxc_GG is not None
            Khxc_GG = Kxc_GG
        else:
            Khxc_GG = self.get_hartree_kernel(pd)
            if Kxc_GG is not None:  # Kxc can be None in the RPA case
                Khxc_GG += Kxc_GG

        chi_wGG = self.invert_dyson(chiks_wGG, Khxc_GG)

        print('\nFinished calculating component', file=self.cfd)
        print('---------------', flush=True, file=self.cfd)

        return pd, wd, chiks_wGG, chi_wGG

    def get_xc_kernel(self, spincomponent, pd, **ignored):
        return self.fxc(spincomponent, pd, txt=self.cfd)

    def get_hartree_kernel(self, pd):
        """Calculate the Hartree kernel"""
        Kbare_G = get_coulomb_kernel(pd, self.gs.kd.N_c)
        vsqrt_G = Kbare_G ** 0.5
        Kh_GG = np.eye(len(vsqrt_G)) * vsqrt_G * vsqrt_G[:, np.newaxis]

        return Kh_GG

    def write_timer(self):
        """Write timer to call-output file and initiate a new."""
        self.timer.write(self.cfd)
        self.timer = Timer()

        # Update all other class instance timers
        self.chiks.timer = self.timer
        self.chiks.integrator.timer = self.timer
        self.chiks.kspair.timer = self.timer
        self.chiks.pme.timer = self.timer
        self.fxc.timer = self.timer

    def calculate_ks_component(self, spincomponent, q_c, wd, txt=None):
        """Calculate a single component of the Kohn-Sham susceptibility tensor.

        Parameters
        ----------
        spincomponent, q_c : see gpaw.response.chiks, gpaw.response.kslrf
        wd : see calculate_component

        Returns
        -------
        pd : PWDescriptor
            see gpaw.response.chiks, gpaw.response.kslrf
        chiks_wGG : ndarray
            The process' block of the Kohn-Sham susceptibility component
        """
        # ChiKS calculates the susceptibility distributed over plane waves
        pd, chiks_wGG = self.chiks.calculate(q_c, wd, txt=txt,
                                             spincomponent=spincomponent)

        # Redistribute memory, so each block has its own frequencies, but all
        # plane waves (for easy invertion of the Dyson-like equation)
        chiks_wGG = self.chiks.distribute_frequencies(chiks_wGG)

        return pd, chiks_wGG

    @timer('Invert dyson-like equation')
    def invert_dyson(self, chiks_wGG, Khxc_GG):
        """Invert the Dyson-like equation:

        chi = chi_ks + chi_ks Khxc chi
        """
        print('Inverting Dyson-like equation', flush=True, file=self.cfd)
        chi_wGG = np.empty_like(chiks_wGG)
        for w, chiks_GG in enumerate(chiks_wGG):
            chi_GG = invert_dyson_single_frequency(chiks_GG, Khxc_GG)

            chi_wGG[w] = chi_GG

        return chi_wGG

    def collect(self, a_w):
        """Collect frequencies from all blocks"""
        return self.blocks1d.collect(a_w)

    def gather(self, A_wGG, wd):
        """Gather a full susceptibility array to root."""
        # Allocate arrays to gather (all need to be the same shape)
        blocks1d = self.blocks1d
        shape = (blocks1d.blocksize,) + A_wGG.shape[1:]
        tmp_wGG = np.empty(shape, dtype=A_wGG.dtype)
        tmp_wGG[:blocks1d.nlocal] = A_wGG

        # Allocate array for the gathered data
        if self.world.rank == 0:
            # Make room for all frequencies
            Npadded = blocks1d.blocksize * blocks1d.blockcomm.size
            shape = (Npadded,) + A_wGG.shape[1:]
            allA_wGG = np.empty(shape, dtype=A_wGG.dtype)
        else:
            allA_wGG = None

        self.world.gather(tmp_wGG, 0, allA_wGG)

        # Return array for w indeces on frequency grid
        if allA_wGG is not None:
            allA_wGG = allA_wGG[:len(wd), :, :]

        return allA_wGG

    def close(self):
        self.timer.write(self.cfd)
        print('\nClosing, %s' % ctime(), file=self.cfd)
        self.cfd.close()
        print('\nClosing, %s' % ctime(), file=self.fd)
        self.fd.close()


def invert_dyson_single_frequency(chiks_GG, Khxc_GG):
    """Invert single frequency Dyson equation in plane wave basis:

    chi_GG' = chiks_GG + chiks_GG1 Khxc_G1G2 chi_G2G'
    """
    enhancement_GG = np.linalg.inv(np.eye(len(chiks_GG)) -
                                   np.dot(chiks_GG, Khxc_GG))
    chi_GG = np.dot(enhancement_GG, chiks_GG)

    return chi_GG


def get_pw_reduction_map(pd, ecut):
    """Get a mask to reduce the plane wave representation.

    Please remark, that the response code currently works with one q-vector
    at a time, at thus only a single plane wave representation at a time.

    Returns
    -------
    mask_G : nd.array (dtype=bool)
        Mask which reduces the representation
    """
    assert ecut is not None
    ecut /= Hartree
    assert ecut <= pd.ecut

    # List of all plane waves
    G_Gv = np.array([pd.G_Qv[Q] for Q in pd.Q_qG[0]])

    if pd.gammacentered:
        mask_G = ((G_Gv ** 2).sum(axis=1) <= 2 * ecut)
    else:
        mask_G = (((G_Gv + pd.K_qv[0]) ** 2).sum(axis=1) <= 2 * ecut)

    return mask_G


def get_pw_coordinates(pd):
    """Get the reciprocal lattice vector coordinates corresponding to a
    givne plane wave basis.

    Please remark, that the response code currently works with one q-vector
    at a time, at thus only a single plane wave representation at a time.

    Returns
    -------
    G_Gc : nd.array (dtype=int)
        Coordinates on the reciprocal lattice
    """
    # List of all plane waves
    G_Gv = np.array([pd.G_Qv[Q] for Q in pd.Q_qG[0]])

    # Use cell to get coordinates
    B_cv = 2.0 * np.pi * pd.gd.icell_cv
    return np.round(np.dot(G_Gv, np.linalg.inv(B_cv))).astype(int)


def get_inverted_pw_mapping(pd1, pd2):
    """Get the plane wave coefficients mapping GG' of pd1 into -G-G' of pd2"""
    G1_Gc = get_pw_coordinates(pd1)
    G2_Gc = get_pw_coordinates(pd2)

    mG2_G1 = []
    for G1_c in G1_Gc:
        found_match = False
        for G2, G2_c in enumerate(G2_Gc):
            if np.all(G2_c == -G1_c):
                mG2_G1.append(G2)
                found_match = True
                break
        if not found_match:
            raise ValueError('Could not match pd1 and pd2')

    # Set up mapping from GG' to -G-G'
    invmap_GG = tuple(np.meshgrid(mG2_G1, mG2_G1, indexing='ij'))

    return invmap_GG


def symmetrize_reciprocity(pd, A_wGG):
    """In collinear systems without spin-orbit coupling, the plane wave
    susceptibility is reciprocal in the sense that e.g.

    χ_(GG')^(+-)(q, ω) = χ_(-G'-G)^(+-)(-q, ω)

    This method symmetrizes A_wGG in the case where q=0.
    """
    from gpaw.test.response.test_chiks import get_inverted_pw_mapping

    q_c = pd.kd.bzk_kc[0]
    if np.allclose(q_c, 0.):
        invmap_GG = get_inverted_pw_mapping(pd, pd)
        for A_GG in A_wGG:
            tmp_GG = np.zeros_like(A_GG)

            # Symmetrize [χ_(GG')(q, ω) + χ_(-G'-G)(-q, ω)] / 2
            tmp_GG += A_GG
            tmp_GG += A_GG[invmap_GG].T
            A_GG[:] = tmp_GG / 2.


def write_macroscopic_component(omega_w, chiks_w, chi_w, filename, world):
    """Write the spatially averaged dynamic susceptibility."""
    assert isinstance(filename, str)
    if world.rank == 0:
        with Path(filename).open('w') as fd:
            for omega, chiks, chi in zip(omega_w, chiks_w, chi_w):
                print('%.6f, %.6f, %.6f, %.6f, %.6f' %
                      (omega, chiks.real, chiks.imag, chi.real, chi.imag),
                      file=fd)


def read_macroscopic_component(filename):
    """Read a stored macroscopic susceptibility file"""
    d = np.loadtxt(filename, delimiter=',')
    omega_w = d[:, 0]
    chiks_w = np.array(d[:, 1], complex)
    chiks_w.imag = d[:, 2]
    chi_w = np.array(d[:, 3], complex)
    chi_w.imag = d[:, 4]

    return omega_w, chiks_w, chi_w


def write_component(omega_w, G_Gc, chiks_wGG, chi_wGG, filename, world):
    """Write the dynamic susceptibility as a pickle file."""
    assert isinstance(filename, str)
    if world.rank == 0:
        with open(filename, 'wb') as fd:
            pickle.dump((omega_w, G_Gc, chiks_wGG, chi_wGG), fd)


def read_component(filename):
    """Read a stored susceptibility component file"""
    assert isinstance(filename, str)
    with open(filename, 'rb') as fd:
        omega_w, G_Gc, chiks_wGG, chi_wGG = pickle.load(fd)

    return omega_w, G_Gc, chiks_wGG, chi_wGG



r"""
Ascii-art math to LaTeX converter
=================================

Examples:

::

 1+2
 ---
  3

.. math::

  \frac{1+2}{3}

::

 <a|b>

.. math::

  \langle a|b \rangle

::

 /  _ ~ ~    --- a
 | dr 𝜓 𝜓  + >  S
 /     m n   --- ij
             aij

.. math::

  \int d\mathbf{r} \tilde{𝜓}_{m} \tilde{𝜓}_{n}  + \sum^{}_{aij} S^{a}_{ij}
"""

from __future__ import annotations


def prep(lines: list[str], n: int) -> list[str]:
    """Preprocess lines.

    * Remove leading and trailing empty lines.
    * Make all lines have the same length (pad with spaces).
    * Remove spaces from beginning of lines.
    """
    if not lines:
        return [], n
    while lines and not lines[0].strip():
        lines.pop(0)
        if n is not None:
            n -= 1
    while lines and not lines[-1].strip():
        lines.pop()
    if not lines:
        return [], 0
    i = min(len(line) - len(line.lstrip()) for line in lines)
    lines = [line[i:] for line in lines]
    i = max(len(line) for line in lines)
    return [line.ljust(i) for line in lines], n


class ParseError(Exception):
    """Bad ascii-art math."""


def cut(lines: list[str], i1: int, i2: int = None) -> list[str]:
    """Cut out block.

    >>> cut(['012345', 'abcdef'], 1, 3)
    ['12', 'bc']
    """
    index = slice(i1, i2)
    return [line[index] for line in lines]


def block(lines: list[str]) -> dict[int, str]:
    r"""Find superscript/subscript blocks.

    >>> block([' 2  _ ',
    ...        '    k '])
    {0: '2', 3: '\\mathbf{k}'}
    """
    if not lines:
        return {}
    blocks = {}
    i1 = None
    for i in range(len(lines[0])):
        if all(line[i] == ' ' for line in lines):
            if i1 is not None:
                blocks[i1 - 1] = parse(cut(lines, i1, i))
                i1 = None
        else:
            if i1 is None:
                i1 = i
    if i1 is not None:
        blocks[i1 - 1] = parse(cut(lines, i1))
    return blocks


def parse(lines: str | list[str], n: int = None) -> str:
    r"""Parse ascii-art math to LaTeX.

    >>> parse([' /   ~      ',
    ...        ' |dx p  (x) ',
    ...        ' /    ai    '])
    '\\int dx \\tilde{p}_{ai}  (x)'
    >>> parse(['   _ _ ',
    ...        '  ik.r ',
    ...        ' e     '])
    'e^{i\\mathbf{k}\\cdot \\mathbf{r}}'
    """
    if isinstance(lines, str):
        lines = lines.splitlines()
    lines, n = prep(lines, n)

    if not not False:
        print('*********************************************')
        print(n)
        print('\n'.join(lines))

    if not lines:
        return ''
    if n is None:
        N = max((len(line.replace(' ', '')), n)
                for n, line in enumerate(lines))[1]
        for n in [N] + [n for n in range(len(lines)) if n != N]:
            try:
                latex = parse(lines, n)
            except ParseError:
                continue
            return latex
        raise ParseError('Could not parse\n\n' + '    \n'.join(lines))

    line = lines[n]
    i1 = line.find('--')
    if i1 != -1:
        i2 = len(line) - len(line[i1:].lstrip('-'))
        p1 = parse(cut(lines, 0, i1), n)
        p2 = parse(cut(lines[:n], i1, i2))
        p3 = parse(cut(lines[n + 1:], i1, i2))
        p4 = parse(cut(lines, i2), n)
        return rf'{p1} \frac{{{p2}}}{{{p3}}} {p4}'.strip()

    i = line.find('>')
    if i != -1 and n > 0 and lines[n - 1][i] == '-':
        line1 = lines[n - 1]
        i2 = len(line1) - len(line1[i:].lstrip('-'))
        p1 = parse(cut(lines, 0, i), n)
        p2 = parse(cut(lines[:n - 1], i, i2))
        p3 = parse(cut(lines[n + 2:], i, i2))
        p4 = parse(cut(lines, i2), n)
        return rf'{p1} \sum^{{{p2}}}_{{{p3}}} {p4}'.strip()

    i = line.find('|')
    if i != -1:
        if n > 0 and lines[n - 1][i] == '/':
            p1 = parse(cut(lines, 0, i), n)
            p2 = parse(cut(lines, i + 1), n)
            return rf'{p1} \int {p2}'.strip()
        i1 = line.find('<')
        i2 = line.find('>')
        if i1 != -1 and i1 <= i and i2 != -1 and i2 >= i:
            p1 = parse(cut(lines, 0, i1), n)
            p2 = parse(cut(lines, i1 + 1, i), n)
            p3 = parse(cut(lines, i + 1, i2), n)
            p4 = parse(cut(lines, i2 + 1), n)
            return rf'{p1} \langle {p2}|{p3} \rangle {p4}'.strip()

    hats = {}
    if n > 0:
        new = []
        for i, c in enumerate(lines[n - 1]):
            if c in '^~_':
                hats[i] = c
                c = ' '
            new.append(c)
        lines[n - 1] = ''.join(new)

    superscripts = block(lines[:n])
    subscripts = block(lines[n + 1:])

    latex = []
    for i, c in enumerate(line):

        if i in hats:
            hat = {'^': 'hat', '~': 'tilde', '_': 'mathbf'}[hats[i]]
            c = rf'\{hat}{{{c}}}'
        sup = superscripts.pop(i, None)
        if sup:
            c = rf'{c}^{{{sup}}}'
        sub = subscripts.pop(i, None)
        if sub:
            c = rf'{c}_{{{sub}}}'
        c = {'.': r'\cdot '}.get(c, c)
        latex.append(c)

    if superscripts or subscripts:
        raise ParseError(f'suoer={superscripts}, sub={subscripts}')

    result = ''.join(latex).strip()

    for sequence, replacement in [
        ('->', r'\rightarrow'),
        ('<-', r'\leftarrow')]:
        result = result.replace(sequence, replacement)

    return result


def autodoc_process_docstring(lines):
    """Hook-function for Sphinx."""
    blocks = []
    for i1, line in enumerate(lines):
        if line.endswith(':::'):
            for i2, line in enumerate(lines[i1 + 2:], i1 + 2):
                if not line:
                    break
            else:
                i2 += 1
            blocks.append((i1, i2))
    for i1, i2 in reversed(blocks):
        latex = parse(lines[i1 + 1:i2])
        line = f'.. math:: {latex}'
        if lines[i1].strip() == ':::':
            lines[i1:i2] = [line]
        else:
            lines[i1:i2] = [lines[i1][:-2], '', line]


def test_examples():
    """Test examples from module docstring."""
    lines = __doc__.replace('\n::', '\n:::').splitlines()
    autodoc_process_docstring(lines)
    for example in '\n'.join(lines).split('.. math:: ')[1:]:
        line1, *lines = example.splitlines()
        line2 = lines[3].strip()
        assert line1 == line2


def main():
    import sys
    import argparse
    import importlib
    parser = argparse.ArgumentParser(
        description='Parse docstring with ascii-art math.')
    parser.add_argument(
        'thing',
        nargs='?',
        help='Name of module, class, method or '
        'function.  Examples: "module.submodule", '
        '"module.Class", "module.Class.method", '
        '"module.function".  Will read from stdin if not given.')
    args = parser.parse_args()
    if args.thing is None:
        lines = sys.stdin.read().splitlines()
        print(parse(lines))
    else:
        parts = args.thing.split('.')
        for i, part in enumerate(parts):
            if not part.islower():
                mod = importlib.import_module('.'.join(parts[:i]))
                break
        else:
            # no break
            try:
                mod = importlib.import_module('.'.join(parts))
                i += 1
            except ImportError:
                mod = importlib.import_module('.'.join(parts[:-1]))
        thing = mod
        for part in parts[i:]:
            thing = getattr(thing, part)
        lines = thing.__doc__.splitlines()
        autodoc_process_docstring(lines)
        print('\n'.join(lines))


if __name__ == '__main__':
    main()



# web-page: Fe_bands.png
import numpy as np
import pylab as pl
from gpaw import GPAW
from gpaw.spinorbit import soc_eigenstates

calc = GPAW('Fe_bands.gpw', txt=None)

x = np.loadtxt('Fe_kpath.dat')
X = np.loadtxt('Fe_highsym.dat')

e_skn = np.array([[calc.get_eigenvalues(kpt=k, spin=s)
                   for k in range(len(calc.get_ibz_k_points()))]
                  for s in range(2)])
e_snk = np.swapaxes(e_skn, 1, 2)
ef = GPAW('Fe_gs.gpw').get_fermi_level()
e_snk -= ef
for e_k in e_snk[0]:
    pl.plot(x, e_k, '--', c='0.5')
for e_k in e_snk[1]:
    pl.plot(x, e_k, '--', c='0.5')

soc = soc_eigenstates(calc)
e_nk = soc.eigenvalues().T
s_knv = soc.spin_projections()

e_nk -= soc.fermi_level
s_nk = (s_knv[:, :, 2].T + 1.0) / 2.0

pl.xticks(X, [r'$\Gamma$', '(010)   H   (001)', r'$\Gamma$'], size=20)
pl.yticks(size=20)
for i in range(len(X))[1:-1]:
    pl.plot(2 * [X[i]], [1.1 * np.min(e_nk), 1.1 * np.max(e_nk)],
            c='0.5', linewidth=0.5)

pl.scatter(np.tile(x, len(e_nk)), e_nk.reshape(-1),
           c=s_nk.reshape(-1),
           edgecolor=pl.get_cmap('jet')(s_nk.reshape(-1)),
           s=5,
           marker='+')
pl.plot([0, x[-1]], 2 * [0.0], '-', c='0.5')

pl.ylabel(r'$\varepsilon_n(k)$ [eV]', size=24)
pl.axis([0, x[-1], -0.5, 0.5])
# pl.show()
pl.savefig('Fe_bands.png')



import numpy as np

from ase.io import read
from ase.dft.kpoints import monkhorst_pack

from gpaw import GPAW, ConvergenceError
from gpaw import PW
from gpaw.mixer import Mixer

# the system does not converge with 3 diis steps, but converges with 2, 4, ...

bulk = read('na2o4.xyz')
bulk.set_cell([4.31110615, 5.60646198, 3.48126881],
              scale_atoms=False)
magmoms = [0.0 for n in range(len(bulk))]
for n, a in enumerate(bulk):
    if a.symbol == 'O':
        magmoms[n] = 0.5
bulk.set_initial_magnetic_moments(magmoms)

bulk.pbc = (True, True, True)

nk = 4
kpts = monkhorst_pack((nk, nk, nk))
kshift = 1.0 / (2 * nk)
kpts += np.array([kshift, kshift, kshift])

calc = GPAW(
    mode=PW(),
    xc='PBE',
    kpts=kpts,
    parallel={'band': 1},
    eigensolver='cg',
    txt='na2o4.txt')

bulk.calc = calc

try:
    bulk.get_potential_energy()
except ConvergenceError:
    pass

assert not calc.scf.converged

del bulk.calc

calc1 = GPAW(
    mode=PW(),
    xc='PBE',
    kpts=kpts,
    parallel={'band': 1},
    eigensolver='cg',
    txt='na2o4.txt')

calc1.set(mixer=Mixer(0.10, 2))
calc1.set(txt='na2o4_m1.txt')
bulk.calc = calc1
bulk.get_potential_energy()



from ase.io import read
from gpaw import GPAW
import gpaw.dscf as dscf
from gpaw.mixer import MixerSum
from numpy import reshape, dot

filename = 'excited'

c_mol = GPAW(nbands=8, h=0.2, xc='RPBE', kpts=(4, 6, 1),
             eigensolver='cg', spinpol=True,
             convergence={'bands': -2}, txt='CO.txt')

calc = GPAW(nbands=120, h=0.2, xc='RPBE', kpts=(4, 6, 1),
            setups={'Pt': '10'},
            eigensolver='cg', spinpol=True,
            mixer=MixerSum(nmaxold=5, beta=0.1, weight=100),
            convergence={'eigenstates': 1.0e-4,
                         'bands': -10}, txt=filename + '.txt')

# Import Slab with relaxed CO
slab = read('CO_Pt.traj')
pos = slab.get_positions()
slab.calc = calc
E_gs = slab.get_potential_energy()

# Molecule
molecule = slab.copy()
del molecule[:-2]
molecule.calc = c_mol

del slab[-2:]

d_0 = pos[-1][2] - pos[-2][2]
z_0 = 16 * pos[-1][2] / (12 + 16) + 12 * pos[-2][2] / (12 + 16)

ds = [d_0 - 0.02 + i * 0.02 for i in range(3)]

E_es = []

for d in ds:
    molecule[0].position = (pos[12][0], pos[12][1], z_0 - 16 * d / (12 + 16))
    molecule[1].position = (pos[13][0], pos[13][1], z_0 + 12 * d / (12 + 16))
    molecule.get_potential_energy()

    # Find band corresponding to lumo
    lumo = c_mol.get_pseudo_wave_function(band=6, kpt=0, spin=1)
    lumo = reshape(lumo, -1)

    wf1_k = [c_mol.get_pseudo_wave_function(band=5, kpt=k, spin=1)
             for k in range(len(c_mol.wfs.kd.weight_k))]
    wf2_k = [c_mol.get_pseudo_wave_function(band=6, kpt=k, spin=1)
             for k in range(len(c_mol.wfs.kd.weight_k))]

    band_k = []
    for k in range(len(c_mol.wfs.kd.weight_k)):
        wf1 = reshape(wf1_k[k], -1)
        wf2 = reshape(wf2_k[k], -1)
        p1 = abs(dot(wf1, lumo))
        p2 = abs(dot(wf2, lumo))
        if p1 > p2:
            band_k.append(5)
        else:
            band_k.append(6)
        print('Kpt', k, p1, p2, 'band', band_k[-1])

    # Lumo wavefunction
    wf_u = [kpt.psit_nG[band_k[kpt.k]] for kpt in c_mol.wfs.kpt_u]

    # Lumo projector overlaps
    mol = [12, 13]
    p_uai = [dict([(mol[a], P_ni[band_k[kpt.k]])
                   for a, P_ni in kpt.P_ani.items()])
             for kpt in c_mol.wfs.kpt_u]

    # Slab with adsorbed molecule
    slab.extend(molecule)
    orbital = dscf.AEOrbital(calc, wf_u, p_uai)
    dscf.dscf_calculation(calc, [[1.0, orbital, 1]], slab)
    E_es.append(slab.get_potential_energy())

    del slab[12:]

F_d = (E_es[0] - E_es[2]) / 0.02 / 2
assert abs(E_es[1] - (E_gs + 3.9)) < 0.1
assert abs(F_d - 3.8) < 0.2



# web-page: W_r.svg
import numpy as np
import matplotlib.pyplot as plt
from qeh import Heterostructure


thick_MoS2 = 6.2926
thick_WSe2 = 6.718

d_MoS2_WSe2 = (thick_MoS2 + thick_WSe2) / 2
inter_mass = 0.244

HS = Heterostructure(structure=['MoS2_int', 'WSe2_int'],
                     d=[d_MoS2_WSe2],
                     qmax=None,
                     wmax=0,
                     d0=thick_WSe2)

hl_array = np.array([0., 0., 1., 0.])
el_array = np.array([1., 0., 0., 0.])


# Getting the interlayer exciton screened interaction on a real grid
r, W_r = HS.get_exciton_screened_potential_r(
    r_array=np.linspace(1e-1, 30, 1000),
    e_distr=el_array,
    h_distr=hl_array)

plt.plot(r, W_r, '-g')
plt.title(r'Screened Interaction Energy')
plt.xlabel(r'${\bf r}$ (Ang)', fontsize=20)
plt.ylabel(r'$W({\bf r})$ (Ha)', fontsize=20)
plt.savefig('W_r.svg')
plt.show()

ee, ev = HS.get_exciton_binding_energies(eff_mass=inter_mass,
                                         e_distr=el_array,
                                         h_distr=hl_array)

print('The interlayer exciton binding energy is:', -ee[0])



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ase.units import Bohr
from gpaw.core.atom_arrays import AtomArrays
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.setup import Setups
from gpaw.spline import Spline
from gpaw.typing import Array3D, ArrayLike2D, Vector
from gpaw.spherical_harmonics import Y

if TYPE_CHECKING:
    from gpaw.new.calculation import DFTCalculation


class Densities:
    def __init__(self,
                 nt_sR: UniformGridFunctions,
                 D_asii: AtomArrays,
                 fracpos_ac: ArrayLike2D,
                 setups: Setups):
        self.nt_sR = nt_sR
        self.D_asii = D_asii
        self.fracpos_ac = fracpos_ac
        self.setups = setups

    @classmethod
    def from_calculation(cls, calculation: DFTCalculation):
        density = calculation.state.density
        return cls(density.nt_sR,
                   density.D_asii,
                   calculation.fracpos_ac,
                   calculation.setups)

    def pseudo_densities(self,
                         grid_spacing: float = None,  # Ang
                         ) -> UniformGridFunctions:
        nt_sR = self._pseudo_densities(grid_spacing)
        return nt_sR.scaled(Bohr, Bohr**-3)

    def _pseudo_densities(self,
                          grid_spacing: float = None,  # Ang
                          grid_refinement: int = None,
                          ) -> UniformGridFunctions:
        nt_sR = self.nt_sR.to_pbc_grid()
        grid = nt_sR.desc
        if grid_spacing is not None:
            grid = grid.uniform_grid_with_grid_spacing(
                grid_spacing / Bohr)
        elif grid_refinement is not None and grid_refinement > 1:
            grid = grid.new(size=grid.size * grid_refinement)
        else:
            return nt_sR

        return nt_sR.interpolate(grid=grid)

    def all_electron_densities(self,
                               *,
                               grid_spacing: float = None,  # Ang
                               grid_refinement: int = None,
                               ) -> UniformGridFunctions:
        n_sR = self._pseudo_densities(grid_spacing, grid_refinement)

        splines = {}
        for R_v, setup, D_sii in zip(self.fracpos_ac @ n_sR.desc.cell_cv,
                                     self.setups,
                                     self.D_asii.values()):
            if setup not in splines:
                phi_j, phit_j, nc, nct = setup.get_partial_waves()[:4]
                rcut = max(setup.rcut_j)
                splines[setup] = (rcut, phi_j, phit_j)
            rcut, phi_j, phit_j = splines[setup]
            add(R_v, n_sR, phi_j, phit_j, rcut, D_sii)

        return n_sR.scaled(Bohr, Bohr**-3)


def add(R_v: Vector,
        a_sR: UniformGridFunctions,
        phi_j: list[Spline],
        phit_j: list[Spline],
        rcut: float,
        D_sii: Array3D):
    ug = a_sR.desc
    R_Rv = ug.xyz()
    lmax = max(phi.l for phi in phi_j)
    start_c = 0 - ug.pbc
    stop_c = 1 + ug.pbc
    for u0 in range(start_c[0], stop_c[0]):
        for u1 in range(start_c[1], stop_c[1]):
            for u2 in range(start_c[2], stop_c[2]):
                d_Rv = R_Rv - (R_v + (u0, u1, u2) @ ug.cell_cv)
                d_R = (d_Rv**2).sum(3)**0.5
                mask_R = d_R < rcut
                if not mask_R.any():
                    continue
                d_rv = d_Rv[mask_R]
                d_r = d_R[mask_R]
                Y_Lr = [Y(L, *d_rv.T) for L in range((lmax + 1)**2)]
                phi_jr = [phi.map(d_r) for phi in phi_j]
                phit_jr = [phit.map(d_r) for phit in phit_j]
                l_j = [phi.l for phi in phi_j]
                i1 = 0
                for l1, phi1_r, phit1_r in zip(l_j, phi_jr, phit_jr):
                    i2 = 0
                    i1b = i1 + 2 * l1 + 1
                    D_smi = D_sii[:, i1:i1b]
                    for l2, phi2_r, phit2_r in zip(l_j, phi_jr, phit_jr):
                        i2b = i2 + 2 * l2 + 1
                        D_smm = D_smi[:, :, i2:i2b]
                        a_sr = np.einsum(
                            'smn, mr, nr -> sr',
                            D_smm,
                            Y_Lr[l1**2:(l1 + 1)**2],
                            Y_Lr[l2**2:(l2 + 1)**2]) * (
                            phi1_r * phi2_r - phit1_r * phit2_r)
                        a_sR.data[:, mask_R] += a_sr
                        i2 = i2b
                    i1 = i1b



import asr
from ase.build import bulk


def materials():
    elements = ['Al', 'Si', 'Ti', 'Fe', 'Ni', 'Cu', 'Ag', 'Sb', 'Au']
    return {symbol: bulk(symbol) for symbol in elements}


workflow = asr.totree(materials(), name='material')



import numpy as np
from ase.units import Hartree
from gpaw.directmin.tools import d_matrix


class KSLCAO:

    """
    This class is used to calculate the gradient of the KS functional
    w.r.t rotation parameters in LCAO
    """

    def __init__(self):
        self.name = 'ks'

    def todict(self):
        return {'name': self.name}

    def get_gradients(self, h_mm, c_nm, f_n, evec, evals, kpt, wfs, timer,
                      matrix_exp, representation, ind_up):

        with timer('Construct Gradient Matrix'):
            use_egdecomp = matrix_exp == 'egdecomp'
            
            hc_mn, h_ij, h_ia = self.get_ham_in_mol_orb_representation(
                h_mm, c_nm, f_n, representation, use_egdecomp)
            
            with timer('Residual'):
                error = self.get_residual_error(
                    hc_mn, kpt.S_MM, c_nm, h_ij, f_n, wfs.nvalence)

            if not use_egdecomp:
                if representation == 'u-invar':
                    h_ij = h_ia
                else:
                    if representation == 'full' and wfs.dtype == complex:
                        indl = np.tril_indices(h_ij.shape[0], -1)
                    else:
                        indl = np.tril_indices(h_ij.shape[0])
                    h_ij[indl] = np.inf
                    if representation == 'sparse':
                        h_ij = np.concatenate((h_ij, h_ia), axis=1)
                h_ij = h_ij.ravel()
                h_ij = h_ij[h_ij != np.inf]
        
                ones_mat = np.ones(shape=(len(f_n), len(f_n)))
                anti_occ = f_n * ones_mat - f_n[:, np.newaxis] * ones_mat
                anti_occ = anti_occ[ind_up]
                grad = np.ascontiguousarray(anti_occ * h_ij)

            else:
                h_ij = f_n * h_ij - f_n[:, np.newaxis] * h_ij
                grad = np.ascontiguousarray(h_ij)

                with timer('Use Eigendecomposition'):
                    grad = self.get_exact_gradient_matrix(grad, evec, evals)

                grad = grad[ind_up]

        if wfs.dtype == float:
            grad = grad.real

        return 2.0 * grad, error

    def get_ham_in_mol_orb_representation(self, h_mm, c_nm, f_n,
                                          representation, full_ham):

        """
        H = (C_nM @ H_MM @ C_nM.T.conj()).conj()
        for sparse and u_inv representation we calculate
        H_ij and H_ia, where i,j -- occupied and a - virtual
        H_ij is really needed only to calculate the residual later
        but it is not needed for u-invar representation.
        When matrix exp calculated using egdecomp method
        we need the whole matrix H though

        :return: H@C_nM[:occ].T, H_ij, H_ia or
                 H@C_nM.T, H, H_ia
        """

        occ = sum(f_n > 1.0e-10)
        if representation in ['sparse', 'u-invar'] \
                and not full_ham:
            hc_mn = h_mm.conj() @ c_nm[:occ].T
            h_ij = hc_mn.T.conj() @ c_nm[:occ].T
            h_ia = hc_mn.T.conj() @ c_nm[occ:].T
        else:
            hc_mn = h_mm.conj() @ c_nm.T
            h_ij = c_nm.conj() @ hc_mn
            h_ia = h_ij[:occ][:, occ:]

        return hc_mn, h_ij, h_ia
    
    def get_residual_error(self, hc_mn, S_MM, c_nm, h_ij, f_n, nvalence):
        """
        Calculate residual error of KS equations
        """
        occ = sum(f_n > 1.0e-10)
        hc_mn = hc_mn[:, :occ] - \
            S_MM.conj() @ c_nm[:occ].T @ h_ij[:occ, :occ]
        norm = sum(hc_mn.conj() * hc_mn * f_n[:occ])
        error = sum(norm.real) * Hartree ** 2 / nvalence
        
        return error

    def get_exact_gradient_matrix(self, h_ij, evec, evals):
 
        """
        Given eigendecomposition of A
        calculate exact gradient matrix
        eq.(14), (39)-(41) from
        arXiv:2101.12597 [physics.comp-ph]
        Comput. Phys. Commun. 267, 108047 (2021).
        :return: gradient matrix
        """

        grad = evec.T.conj() @ h_ij @ evec
        grad = grad * d_matrix(evals)
        grad = evec @ grad @ evec.T.conj()
        for i in range(grad.shape[0]):
            grad[i][i] *= 0.5
        
        return grad



"""Test hyperfine parameter CLI."""
import pytest
import numpy as np

from gpaw.hyperfine import main


@pytest.mark.parametrize('option', ['', '-x', '-d'])
@pytest.mark.serial
def test_hyperfine_cli(gpw_files, capsys, option):
    file = gpw_files['o2_pw']
    main(f'{file} -g O:-0.757 -u MHz {option}'.split())
    out = capsys.readouterr().out
    print(out)
    lines = [line.split() for line in out.splitlines()]
    if option == '-x':
        ref = [0.689, 33.74, -246.58, -246.58, 0.0, 0.0, 0.0]
    elif option == '-d':
        ref = [0.689, -78.01, -78.01, 202.31]
    else:
        ref = [0.689, 202.31, -78.01, -78.01, 0.0, 0.0, 0.0]
    for line in lines[3:5]:
        errors = (np.array([float(x) for x in line[2:]]) - ref)
        assert abs(errors).max() < 0.2
    assert float(lines[7][3]) == 2.0
    assert float(lines[10][1]) == -0.757



"""Submit tests from the test suite that were removed because they were
too long."""

from myqueue.workflow import run


def workflow():
    run(script='H2Al110.py')
    run(script='dscf_CO.py')
    run(script='revtpss_tpss_scf.py')
    run(script='ltt.py')
    run(script='scalapack.py', cores=16)
    run(script='pblacs_oblong.py', cores=64)



"""EXX hydrogen atom.

Compare self-consistent EXX calculation with non self-consistent
EXX calculation on top of LDA.
"""
from ase import Atoms
from ase.units import Ry
from gpaw import GPAW, PW
from gpaw.hybrids.eigenvalues import non_self_consistent_eigenvalues
from gpaw.hybrids.energy import non_self_consistent_energy

atoms = Atoms('H', magmoms=[1.0])
atoms.center(vacuum=5.0)

# Self-consistent calculation:
atoms.calc = GPAW(mode=PW(600),
                  xc='EXX:backend=pw')
eexx = atoms.get_potential_energy() + atoms.calc.get_reference_energy()

# Check energy
eexxref = -1.0 * Ry
assert abs(eexx - eexxref) < 0.001

# ... and eigenvalues
eig1, eig2 = (atoms.calc.get_eigenvalues(spin=spin)[0] for spin in [0, 1])
eigref1 = -1.0 * Ry
eigref2 = ...  # ?
assert abs(eig1 - eigref1) < 0.03
# assert abs(eig2 - eigref2) < 0.03

# LDA:
atoms.calc = GPAW(mode=PW(600),
                  xc='LDA')
atoms.get_potential_energy()

# Check non self-consistent eigenvalues
result = non_self_consistent_eigenvalues(atoms.calc,
                                         'EXX',
                                         snapshot='h-hse-snapshot.json')
eiglda, vlda, vexx = result
eig1b, eig2b = (eiglda - vlda + vexx)[:, 0, 0]
assert abs(eig1b - eig1) < 0.04
assert abs(eig2b - eig2) < 1.1

# ... and energy
energies = non_self_consistent_energy(atoms.calc, 'EXX')
eexxb = energies.sum() + atoms.calc.get_reference_energy()
assert abs(eexxb - eexx) < 0.03



from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
from gpaw import GPAW, PW, FermiDirac

calc = GPAW(mode=PW(600),
            xc='PBE',
            setups={'W': '6'},
            parallel={'band': 1, 'domain': 1},
            occupations=FermiDirac(width=0.01),
            kpts=(15, 15, 1),
            txt='WS2_gs.txt',
            )

a = 3.160
c = 12.0

cell = Hexagonal(symbol='W', latticeconstant={'a': a, 'c': c}).get_cell()
layer = Atoms(symbols='WS2', cell=cell, pbc=(1, 1, 1),
              scaled_positions=[(0, 0, 0),
                                (2 / 3, 1 / 3, 0.3),
                                (2 / 3, 1 / 3, -0.3)])

layer.center(axis=2)
layer.set_pbc(True)
pos = layer.get_positions()
pos[1][2] = pos[0][2] + 3.172 / 2
pos[2][2] = pos[0][2] - 3.172 / 2
layer.set_positions(pos)
layer.calc = calc
layer.get_potential_energy()
calc.write('WS2_gs.gpw')



from gpaw import GPAW, PW
from ase.parallel import parprint
from ase.lattice.compounds import L1_2


def test_xc_qna_force(in_tmp_dir):
    name = 'Cu3Au'
    ecut = 300
    kpts = (2, 2, 2)

    QNA = {'alpha': 2.0,
           'name': 'QNA',
           'stencil': 1,
           'orbital_dependent': False,
           'parameters': {'Au': (0.125, 0.1), 'Cu': (0.0795, 0.005)},
           'setup_name': 'PBE',
           'type': 'qna-gga'}

    atoms = L1_2(['Au', 'Cu'], latticeconstant=3.74)
    # Displace atoms to have non-zero forces in the first place
    atoms[0].position[0] += 0.1

    dx_array = [-0.005, 0.000, 0.005]
    E = []

    for i, dx in enumerate(dx_array):
        calc = GPAW(mode=PW(ecut),
                    experimental={'niter_fixdensity': 2},
                    xc=QNA,
                    kpts=kpts,
                    parallel={'domain': 1},
                    txt=name + '%.0f.txt' % ecut)

        atoms[0].position[0] += dx
        atoms.calc = calc
        E.append(atoms.get_potential_energy(force_consistent=True))
        if i == 1:
            F = atoms.get_forces()[0, 0]
        atoms[0].position[0] -= dx

    F_num = -(E[-1] - E[0]) / (dx_array[-1] - dx_array[0])
    F_err = F_num - F

    parprint('Analytical force = ', F)
    parprint('Numerical  force = ', F_num)
    parprint('Difference       = ', F_err)
    assert abs(F_err) < 0.01, F_err
    eerr = abs(E[-1] - 270.17901094)
    assert eerr < 1e-3, eerr



from gpaw.xc import XC


class XCFunctional:
    def __init__(self, params: dict, ncomponents: int):
        self.xc = XC(params, collinear=(ncomponents < 4))
        self.setup_name = self.xc.get_setup_name()
        self.name = self.xc.name
        self.no_forces = self.name.startswith('GLLB')
        self.type = self.xc.type

    def __str__(self):
        return f'name: {self.name}'

    def calculate(self, density, out) -> float:
        return self.xc.calculate(density.desc._gd, density.data, out.data)

    def calculate_paw_correction(self, setup, d, h):
        return self.xc.calculate_paw_correction(setup, d, h)

    def get_setup_name(self):
        return self.name



import pytest
import numpy as np
from gpaw.mpi import world
from gpaw.atom.generator2 import generate


@pytest.mark.skipif(world.size > 1, reason='Not parallelized')
def test_diamond(in_tmp_dir, add_cwd_to_setup_paths):

    from gpaw.atom.generator import Generator
    from gpaw.atom.configurations import parameters
    from gpaw.atom.basis import BasisMaker

    args = parameters['C']
    generator = Generator('C', 'PBE', configuration='1s2')
    generator.N *= 2
    generator.run(write_xml=True, exx=True, **args)
    bm = BasisMaker(generator, run=False)
    basis = bm.generate(zetacount=4, polarizationcount=2,
                        energysplit=0.0001, rcutmax=12.0)
    basis.write_xml()

    parameters = {'symbol': 'C', 'Z': 6, 'xc': 'PBE',
                  'projectors': '2s,s,2p,p,d', 'radii': [1.2],
                  'scalar_relativistic': False, 'r0': 1.2,
                  'v0': None,
                  'nderiv0': 2, 'pseudize': ('poly', 4),
                  'omega': 0.11}

    gen = generate(**parameters)
    setup = gen.make_paw_setup()
    setup.write_xml()

    from gpaw import GPAW
    from ase import Atoms
    from ase.build import bulk
    from ase.units import Hartree

    atoms = Atoms('C')
    atoms.center(vacuum=3)
    # calc = GPAW(mode=PW(700), xc='HSE06', charge=4)
    # atoms.calc = calc
    # atoms.get_potential_energy()

    calc = GPAW(h=0.12, mode='lcao', basis='qzdp',
                xc='HSE06WIP:backend=ri', charge=4)
    atoms.calc = calc
    atoms.get_potential_energy()

    # TURBOMOLE C^4+ calculation aug-cc-pV5Z
    tm_eigs = [-68.5345, -61.6901]
    tm_e = -32.35980873243 * Hartree

    assert np.allclose(tm_eigs, calc.get_eigenvalues()[:2], atol=0.1)
    assert np.allclose(tm_e, calc.hamiltonian.setups.Eref * Hartree +
                       atoms.get_potential_energy(), atol=0.1)

    atoms = bulk('C', 'diamond')
    atoms.calc = GPAW(kpts={'size': (2, 2, 2), 'gamma': True},
                      mode='lcao', basis='dzp',
                      xc='HSE06WIP:backend=ri')

    # NOTE: HSE06 does not yet work. This is just a placeholder for
    # integration test.
    assert np.allclose(atoms.get_potential_energy(), +3.291392)

    calc = atoms.calc

    setup = calc.density.setups[0]
    # With w=0 the values should agree exactly. For small systems
    # with small paw-cutoffs, the values are very close nevertheless
    # Thus assert with a moderate tolerance.
    assert np.allclose(setup.M_pp,
                       setup.M_wpp[0.11],
                       rtol=1e-2, atol=1e-4)

    assert np.allclose(setup.X_p, setup.X_wp[0.11], atol=1e-5)



from ase.build import bulk
from gpaw.new.ase_interface import GPAW

# Construct bulk iron with FCC lattice
atoms = bulk('Fe', 'fcc', a=3.50)  # V = 10.72
# atoms = bulk('Fe', 'fcc', a=3.577)  # V = 11.44

# Align the magnetic moment in the xy-plane
magmoms = [[1, 0, 0]]
ecut = 600
k = 14

# Construct list of q-vectors
path = atoms.cell.bandpath('GXW', npoints=31)

results = []
for i, q_c in enumerate(path.kpts):
    # Spin-spiral calculations require non-collinear calculations
    # without symmetry or spin-orbit coupling
    calc = GPAW(mode={'name': 'pw',
                      'ecut': ecut,
                      'qspiral': q_c},
                xc='LDA',
                symmetry='off',
                magmoms=magmoms,
                kpts=(k, k, k),
                txt=f'gs-{i:02}.txt')
    atoms.calc = calc
    atoms.get_potential_energy()
    atoms.calc.write(f'gs-{i:02}.gpw')



# web-page: traces-Au111.png

import os
import numpy as np
from matplotlib import pyplot
import ase.io

fig, ax = pyplot.subplots(figsize=(4.6, 2.0))
fig.subplots_adjust(bottom=0.2, top=0.99, left=0.15, right=0.97)

data = np.loadtxt(os.path.join('sjm_traces.out', 'cavity.txt'),
                  delimiter=' ')
ax.plot(data[:, 0], data[:, 1], label='solvent')
data = np.loadtxt(os.path.join('sjm_traces.out', 'background_charge.txt'),
                  delimiter=' ')
ax.plot(data[:, 0], data[:, 1], ':', label='jellium')

# Also add atom dots.
atoms = ase.io.read('Au111.traj')
for atom in atoms:
    ax.plot(atom.z, 0.5, 'k.')

ax.set_xlabel('$z$')
ax.set_ylabel('$xy$-averaged value')
ax.legend()
fig.savefig('traces-Au111.png')



from myqueue.workflow import run


def workflow():
    with run(script='timepropagation_calculate.py', cores=8, tmax='1h'):
        with run(script='timepropagation_continue.py', cores=8, tmax='1h'):
            with run(script='timepropagation_postprocess.py', cores=8):
                run(script='timepropagation_plot.py')

    with run(script='casida_calculate.py', cores=8, tmax='1h'):
        with run(script='casida_postprocess.py', cores=8):
            run(script='casida_plot.py')



from ase.io import read
from gpaw import GPAW

# the system losing magnetic moment

atoms = read('ScSZ.xyz')
atoms.set_cell([[7.307241, 0., 0.],
               [0., 12.656514, 0.],
               [0., 0., 19.]],
               scale_atoms=False)
atoms.center(axis=2)
magmoms = [0.0 for n in range(len(atoms))]
for n, a in enumerate(atoms):
    if a.symbol == 'Ni':
        magmoms[n] = 0.6
atoms.set_initial_magnetic_moments(magmoms)

atoms.pbc = (True, True, False)

atoms.calc = GPAW(h=0.2,
                  kpts=(2, 1, 1),
                  xc='PBE',
                  maxiter=100,
                  txt='ScSZ.txt')

ncpus = 8



def workflow():
    from myqueue.workflow import run
    with run(script='HAl100.py'):
        run(function=go)


def go():
    import sys
    sys.argv = ['', 'HAl100.gpw']
    exec(open('stm.py').read())



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Battery project
"""

# %%
"""
## Day 3 - Equilibrium potential
"""

# %%
"""
Today you will study the LiFePO$_4$ cathode. You will calculate the equilibrium potential and use Bayesian error estimation to quantify how sensitive the calculated equilibrium potential is towards choice of functional. After today you should be able to discuss:

-  The volume change during charge/discharge.

-  The maximum gravimetric and volumetric energy density of a FePO$_4$/C battery assuming the majority of weight and volume will be given by the electrodes.

-  Uncertainty in the calculations.

Some of calculations you will perform today will be tedious to be run in this notebook. You will automatically submit some calculations to the HPC cluster directly from this notebook. When you have to wait for calculations to finish you can get started on addressing the bullet points above.
"""

# %%
"""
## Initialize
"""

# %%
# magic: %matplotlib notebook
import numpy as np
from ase.visualize import view
import matplotlib.pyplot as plt
from ase.io import read, write, Trajectory
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac, Mixer, PW
from ase.dft.bee import BEEFEnsemble
from ase import Atoms

# %%
"""
## FePO$_4$
"""

# %%
"""
First we will construct an atoms object for FePO$_4$. ASE can read files from in a large number of different [formats](https://wiki.fysik.dtu.dk/ase/ase/io/io.html?highlight=read%20formats#file-input-and-output). However, in this case you will build it from scratch using the below information:
"""

# %%
# Positions:
# Fe      2.73015081       1.46880951       4.56541172
# Fe      2.23941067       4.40642872       2.14957739
# Fe      7.20997230       4.40642925       0.26615813
# Fe      7.70070740       1.46880983       2.68199421
# O       1.16033403       1.46881052       3.40240205
# O       3.80867172       4.40642951       0.98654342
# O       8.77981469       4.40642875       1.42923946
# O       6.13142032       1.46881092       3.84509827
# O       4.37288562       1.46880982       0.81812712
# O       0.59764596       4.40643021       3.23442747
# O       5.56702590       4.40642886       4.01346264
# O       9.34268360       1.46880929       1.59716233
# O       1.64001691       0.26061277       1.17298291
# O       3.32931769       5.61463705       3.58882629
# O       8.30013707       3.19826250       3.65857000
# O       6.61076951       2.67698811       1.24272700
# O       8.30013642       5.61459688       3.65856912
# O       6.61076982       0.26063178       1.24272567
# O       1.64001666       2.67700652       1.17298270
# O       3.32931675       3.19822249       3.58882660
# P       0.90585688       1.46880966       1.89272372
# P       4.06363530       4.40642949       4.30853266
# P       9.03398503       4.40642957       2.93877879
# P       5.87676435       1.46881009       0.52297232


# Unit cell:
#            periodic     x          y          z
#   1. axis:    yes    9.94012    0.00000    0.00000
#   2. axis:    yes    0.00000    5.87524    0.00000
#   3. axis:    yes    0.00000    0.00000    4.83157


# %%
"""
You *can* use the cell below as a starting point.
"""

# %%
# fepo4 = Atoms('Fe4O...',
#               positions=[[x0, y0, z0],[x1, y1, z1]...],
#               cell=[x, y, z],
#               pbc=[True, True, True])

# Teacher:
fepo4 = Atoms('Fe4O16P4',
             positions=[[2.73015081, 1.46880951, 4.56541172],
                [2.23941067, 4.40642872, 2.14957739],
                [7.20997230, 4.40642925, 0.26615813],
                [7.70070740, 1.46880983, 2.68199421],
                [1.16033403, 1.46881052, 3.40240205],
                [3.80867172, 4.40642951, 0.98654342],
                [8.77981469, 4.40642875, 1.42923946],
                [6.13142032, 1.46881092, 3.84509827],
                [4.37288562, 1.46880982, 0.81812712],
                [0.59764596, 4.40643021, 3.23442747],
                [5.56702590, 4.40642886, 4.01346264],
                [9.34268360, 1.46880929, 1.59716233],
                [1.64001691, 0.26061277, 1.17298291],
                [3.32931769, 5.61463705, 3.58882629],
                [8.30013707, 3.19826250, 3.65857000],
                [6.61076951, 2.67698811, 1.24272700],
                [8.30013642, 5.61459688, 3.65856912],
                [6.61076982, 0.26063178, 1.24272567],
                [1.64001666, 2.67700652, 1.17298270],
                [3.32931675, 3.19822249, 3.58882660],
                [0.90585688, 1.46880966, 1.89272372],
                [4.06363530, 4.40642949, 4.30853266],
                [9.03398503, 4.40642957, 2.93877879],
                [5.87676435, 1.46881009, 0.52297232]
                        ],
             cell=[9.94012, 5.87524, 4.83157],
             pbc=[1, 1, 1])

# %%
"""
Visualize the structure you have made. Explore the different functions in the visualizer and determine the volume of the cell (`View -> Quick Info`).
"""

# %%
view(fepo4)

# %%
"""
For better convergence of calculations you should specify initial magnetic moments to iron. The iron will in this structure be Fe$^{3+}$ as it donates two *4s* electrons and one *3d* electron to PO$_4$$^{3-}$. What is the magnetic moment of iron? For simplicity you should assume that FePO$_4$ is ferromagnetic.
"""

# %%
# Teacher:
for atom in fepo4:
    if atom.symbol == 'Fe':
        atom.magmom = 5.0  # student: atom.magmom = ?

# %%
"""
Now examine the initial magnetic moments of the system using an [appropriate method](https://wiki.fysik.dtu.dk/ase/ase/atoms.html?highlight=get_initial#list-of-all-methods).
"""

# %%
magmoms = fepo4.get_initial_magnetic_moments()  # student: magmoms = fepo4.xxx()
print(magmoms)

# %%
"""
Write your atoms object to file.
"""

# %%
write('fepo4.traj', fepo4)

# %%
"""
For this calculation you will use the BEEF-vdW functional developed by [Wellendorff et al.](https://journals.aps.org/prb/abstract/10.1103/PhysRevB.85.235149) Although there are better alternatives for calculating the energy of bulk systems, the BEEF-vdW has a build-in ensemble for error estimation of calculated energies. In the set-up of this calculator you will append relevant keyword values into a dictionary, which is inputted the calculator object.
"""

# %%
params_GPAW = {}

# %%
"""
To save computational time while keeping the calculations physically sound, the following should be used:
"""

# %%
params_GPAW['mode']        = PW(500)                      # The used plane wave energy cutoff
params_GPAW['nbands']      = -40                          # The number on empty bands had the system been spin-paired
params_GPAW['kpts']        = {'size': (2, 4, 5),          # The k-point mesh
                              'gamma': True}
params_GPAW['spinpol']     = True                         # Performing spin polarized calculations
params_GPAW['xc']          = 'BEEF-vdW'                   # The used exchange-correlation functional
params_GPAW['occupations'] = FermiDirac(width=0.1,        # The smearing
                                        fixmagmom=True)   # Total magnetic moment fixed to the initial value
params_GPAW['convergence'] = {'eigenstates': 1.0e-4,      # eV^2 / electron
                              'energy':      2.0e-4,      # eV / electron
                              'density':     1.0e-3}
params_GPAW['mixer']       = Mixer(0.1, 5, weight=100.0)  # The mixer used during SCF optimization


# %%
"""
DFT suffers from a so-called self-interaction error. An electron interacts with the system electron density, to which it contributes itself. The error is most pronounced for highly localized orbitals. [Hubbard U correction](https://wiki.fysik.dtu.dk/gpaw/tutorialsexercises/energetics/hubbardu/hubbardu.html) is used to mitigate the self-interaction error of the highly localized *3d*-electrons of Fe. This is done in GPAW using the `setups` keyword.
"""

# %%
params_GPAW['setups']      = {'Fe': ':d,4.3'}             # U=4.3 applied to d orbitals


# %%
"""
Make a GPAW calculator and attach it to the atoms object. Here you will use [get_potential_energy](https://wiki.fysik.dtu.dk/ase/ase/atoms.html#ase.Atoms.get_potential_energy) to start the calculation.
"""

# %%
calc = GPAW(**params_GPAW)
fepo4.calc = calc
epot_fepo4_cell = fepo4.get_potential_energy()
print(epot_fepo4_cell)
write('fepo4_out.traj', fepo4)


# %%
"""
You will use the ensemble capability of the BEEF-vdW functional. You will need this later so you should write it to file so you do not have to start all over again later. Start by obtaining the required data from the calculator, i.e., the individual energy of each term in the BEEF-vdW functional expansion. Get the energy difference compared to BEEF-vdW for 2000 ensemble functionals.
"""

# %%
from ase.dft.bee import BEEFEnsemble

ens = BEEFEnsemble(calc)
dE = ens.get_ensemble_energies(2000)


# %%
"""
Print the energy differences to file. This is not the most efficient way of printing to file but can allow easier subsequent data treatment.
"""

# %%
with paropen('ensemble_fepo4.dat', 'a') as result:
    for e in dE:
        print(e, file=result)


# %%
"""
You now have what you need to make a full script. Make it in the cell below and execute it to make sure the script runs. Once you have made sure the calculation is able to run, stop it by `interupt the kernel`.
"""

# %%
# %%writefile 'fepo4.py'
from ase.parallel import paropen
from ase.io import read, write
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW

# Read in the structure you made and wrote to file above
fepo4 = read('fepo4.traj')

params_GPAW = {...}

# do calculation ...
# BEEF ...
# write ensemble_fepo4.dat file ...

write('fepo4_out.traj', fepo4)


# Teacher:
from ase.parallel import paropen
from ase.io import read, write
from ase.io.trajectory import Trajectory
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW
fepo4 = read('fepo4.traj')

params_GPAW = {}
params_GPAW['mode']        = PW(500)                     #The used plane wave energy cutoff
params_GPAW['nbands']      = -40                           #The number on empty bands had the system been spin-paired
params_GPAW['kpts']        = {'size':  (2,4,5),            #The k-point mesh
                              'gamma': True}
params_GPAW['spinpol']     = True                          #Performing spin polarized calculations
params_GPAW['xc']          = 'BEEF-vdW'                    #The used exchange-correlation functional
params_GPAW['occupations'] = FermiDirac(width = 0.1,      #The smearing
                                        fixmagmom = True)  #Total magnetic moment fixed to the initial value
params_GPAW['convergence'] = {'eigenstates': 1.0e-4,       #eV^2 / electron
                              'energy':      2.0e-4,       #eV / electron
                              'density':     1.0e-3,}
params_GPAW['mixer']       = Mixer(0.1, 5, weight=100.0)   #The mixer used during SCF optimization
params_GPAW['setups']      = {'Fe': ':d,4.3'}              #U=4.3 applied to d orbitals

calc = GPAW(**params_GPAW)
fepo4.calc = calc
epot_fepo4_cell=fepo4.get_potential_energy()
print('E_Pot=', epot_fepo4_cell)

write('fepo4_out.traj', fepo4)

ens = BEEFEnsemble(calc)
dE = ens.get_ensemble_energies(2000)

with paropen('ensemble_fepo4.dat', 'a') as result:
    for e in dE:
        print(e, file=result)

# %%
"""
Uncomment the `%%writefile` line and execute the cell again and submit the calculation to the HPC cluster. The calculation should take around 10 minutes.
"""

# %%
# magic: !mq submit fepo4.py -R 8:1h  # submits the calculation to 8 cores, 1 hour

# %%
"""
Run the below cell to examine the status of your calculation. If no output is returned, the calculation has either finished or failed.
"""

# %%
# magic: !mq ls

# %%
"""
Once the calculation begins, you can run the cells below to open the error log and output of the calculation in a new window. This can be done while the calculation is running.
"""

# %%
# Error log
# magic: !cat "$(ls -t fepo4.py.*err | head -1)"

# %%
# Output
# magic: !cat "$(ls -t fepo4.py.*out | head -1)"

# %%
"""
Once the calculation has finished, load in the result. You can skip past this cell and return later.
"""

# %%
try:
    fepo4 = read('fepo4_out.traj')
    print('Calculation finished')
except FileNotFoundError:
    print('Calculation has not yet finished')

# %%
"""
## LiFePO$_4$


"""

# %%
"""
You will now do similar for LiFePO$_4$. In this case you will load in a template structure called `lifepo4_wo_li.traj` missing only the Li atoms. It is located in the resources folder.
"""

# %%
lifepo4_wo_li = read('lifepo4_wo_li.traj')

# %%
"""
Visualize the structure.
"""

# %%
view(lifepo4_wo_li)

# %%
"""
You should now add Li into the structure using the fractional coordinates below:
"""

# %%
# Li  0    0    0
# Li  0    0.5  0
# Li  0.5  0.5  0.5
# Li  0.5  0    0.5


# %%
"""
Add Li atoms into the structure, e.g., by following the example in [this ASE tutorial](https://wiki.fysik.dtu.dk/ase/gettingstarted/manipulating_atoms/manipulating_atoms.html?highlight=set_cell#manipulating-atoms).
"""

# %%
from numpy import identity
from ase import Atom

cell = lifepo4_wo_li.get_cell()

# ...

# lifepo4 = lifepo4_wo_li.copy()

# Teacher:
from numpy import identity
from ase import Atom

lifepo4 = lifepo4_wo_li.copy()
cell = lifepo4.get_cell()
xyzcell = identity(3)
lifepo4.set_cell(xyzcell, scale_atoms=True)  # Set the unit cell and rescale
lifepo4.append(Atom('Li', (0, 0, 0)))
lifepo4.append(Atom('Li', (0, 0.5, 0)))
lifepo4.append(Atom('Li', (0.5, 0.5, 0.5)))
lifepo4.append(Atom('Li', (0.5, 0, 0.5)))
lifepo4.set_cell(cell, scale_atoms=True)

# %%
"""
Visualize the structure with added Li.
"""

# %%
view(lifepo4)

# %%
"""
Ensure that the magnetic moments are as they should be, once again assuming ferromagnetism for simplicity.
"""

# %%
# ...

# teacher
print(lifepo4.get_initial_magnetic_moments())

# %%
"""
At this point you should save your structure by writing it to a trajectory file.
"""

# %%
write('lifepo4.traj', lifepo4)

# %%
"""
You should now calculate the potential energy of this sytem using the method and same calculational parameters as for FePO$_4$ above. Make a full script in the cell below similar to what you did above for FePO$_4$ and make sure that it runs.
"""

# %%
# %% writefile 'lifepo4.py'
from ase.parallel import paropen
from ase.io import read, write
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW

# Read in the structure you made and wrote to file above
lifepo4 = read('lifepo4.traj')

params_GPAW = {...}

# ...
# ...
# ...

# write('lifepo4_out.traj', lifepo4)

# teacher
from ase.parallel import paropen
from ase.io import read
from ase.io.trajectory import Trajectory
from ase.dft.bee import BEEFEnsemble
from gpaw import GPAW, FermiDirac, Mixer, PW

#Read in the structure you made and wrote to file above
lifepo4 = read('lifepo4.traj')

params_GPAW = {}
params_GPAW['mode']        = PW(500)                     #The used plane wave energy cutoff
params_GPAW['nbands']      = -40                           #The number on empty bands had the system been spin-paired
params_GPAW['kpts']        = {'size':  (2,4,5),            #The k-point mesh
                              'gamma': True}
params_GPAW['spinpol']     = True                          #Performing spin polarized calculations
params_GPAW['xc']          = 'BEEF-vdW'                    #The used exchange-correlation functional
params_GPAW['occupations'] = FermiDirac(width = 0.1,      #The smearing
                                        fixmagmom = True)  #Total magnetic moment fixed to the initial value
params_GPAW['convergence'] = {'eigenstates': 1.0e-4,       #eV^2 / electron
                              'energy':      2.0e-4,       #eV / electron
                              'density':     1.0e-3,}
params_GPAW['mixer']       = Mixer(0.1, 5, weight=100.0)   #The mixer used during SCF optimization
params_GPAW['setups']      = {'Fe': ':d,4.3'}              #U=4.3 applied to d orbitals

calc = GPAW(**params_GPAW)
lifepo4.calc = calc
epot_lifepo4_cell=lifepo4.get_potential_energy()
print('E_Pot=', epot_lifepo4_cell)

traj=Trajectory('lifepo4_out.traj', mode='w', atoms=lifepo4)
traj.write()

ens = BEEFEnsemble(calc)
dE = ens.get_ensemble_energies(2000)
result = paropen('ensemble_lifepo4.dat','a')
for e in dE:
    print(e, file=result)
result.close()

# %%
"""
If the code runs, submit to the HPC cluster as you did above. The calculation takes approximately 10 minutes.
"""

# %%
# magic: !mq submit lifepo4.py -R 8:1h  # submits the calculation to 8 cores, 1 hour

# %%
"""
Run the below cell to examine the status of your calculation. If no output is returned, the calculation has either finished or failed.
"""

# %%
# magic: !mq ls

# %%
"""
Once the calculation begins, you can run the cells below to open the error log and output of the calculation in a new window.
"""

# %%
# Error log
# magic: !cat "$(ls -t lifepo4.py.*err | head -1)"

# %%
# Output
# magic: !cat "$(ls -t lifepo4.py.*out | head -1)"

# %%
"""
When calculation has finished, load in the result. You can skip past this cell and return later.
"""

# %%
try:
    lifepo4=read('lifepo4_out.traj')
    print('Calculation finished')
except FileNotFoundError:
    print('Calculation has not yet finished')

# %%
"""
### Li metal

We use a Li metal reference to calculate the equilibrium potential. On exercise day 2 you used a Li metal reference to calculate the intercalation energy in the graphite anode. The approach is similar here. Just read in the result of the calculation with DFTD3 and attach a new calulator to it.
"""

# %%
from ase import Atoms
from gpaw import GPAW, FermiDirac, PW
from ase.io import read

li_metal = read('Li-metal-DFTD3.traj')  # Change file name accordingly

calc = GPAW(mode=PW(500),
            kpts=(8, 8, 8),
            occupations=FermiDirac(0.15),
            nbands=-10,
            txt=None,
            xc='BEEF-vdW')

li_metal.calc = calc
li_metal.get_potential_energy()
li_metal.write('li_metal.traj')

# %%
"""
Now calculate the ensemble in the same way as for FePO$_4$ and LiFePO$_4$.
"""

# %%
ens = BEEFEnsemble(calc)
li_metal_ens_cell= ens.get_ensemble_energies(2000)
with paropen('ensemble_li_metal.dat', 'a') as result:
    for e in li_metal_ens_cell:
        print(e, file=result)

# %%
"""
## Calculate equilibrium potential and uncertainty
"""

# %%
"""
You can now calculate the equilibrium potential for the case of a FePO$_4$/Li metal battery from the intercallation energy of Li in FePO$_4$. For simplicity, use that assumption that all vibrational energies and entropic terms cancel each other. You should now have completed all submitted calculations before you proceed.
"""

# %%
"""
The calculated energies are for the full cells. Convert them to the energy per formula unit. The [len(...)](https://docs.python.org/3.6/library/functions.html#len) function can be quite helpful for this.
"""

# %%
epot_fepo4_cell=fepo4.get_potential_energy()
epot_lifepo4_cell=lifepo4.get_potential_energy()
epot_li_metal_cell=li_metal.get_potential_energy()
print('epot_fepo4_cell =', epot_fepo4_cell)
print('epot_lifepo4_cell =', epot_lifepo4_cell)
print('epot_li_metal_cell =', epot_li_metal_cell)

# %%
epot_fepo4 = epot_fepo4_cell / len(fepo4) * 6  # student: epot_fepo4 = ...
epot_lifepo4 = epot_lifepo4_cell / len(lifepo4) * 7  # student: epot_lifepo4 = ...
epot_li_metal = epot_li_metal_cell / len(li_metal)  # student: epot_li_metal = ...
# print(epot_fepo4, ...)

# %%
"""
No calculate the equilibrium potential under the assumption that it is given by $V_{eq} = \Delta U /e $, where $U$ is the electronic potential energy of the system and $e$ is the number of electrons transfered.
"""

# %%
# V_eq = ...

# teacher
V_eq = epot_lifepo4 - epot_fepo4 - epot_li_metal
print(V_eq)

# %%
"""
You will now calculate the error estimate for the Li intercallation energy in FePO$_4$ using the BEEF ensemble results. Start by loading in the files. Wait a few minutes and rerun the cell if the number is not 2000 for all of them.
"""

# %%
fepo4_ens_cell = np.genfromtxt('ensemble_fepo4.dat', max_rows=2000)
lifepo4_ens_cell = np.genfromtxt('ensemble_lifepo4.dat', max_rows=2000)

print('number of functionals in ensemble=', len(fepo4_ens_cell))
print('number of functionals in ensemble=', len(lifepo4_ens_cell))
print('number of functionals in ensemble=', len(li_metal_ens_cell))

# %%
"""
Note that these are energies per cell and not per formula unit. Convert them as you did the potential energies above. Note that you are now performing the operation on a list of length 2000 and not a single float value as before.
"""

# %%
# fepo4_ens = fepo4_ens_cell / ...
# ...
# ...

# teacher
fepo4_ens = fepo4_ens_cell / len(fepo4) * 6
lifepo4_ens = lifepo4_ens_cell / len(lifepo4) * 7
li_metal_ens = li_metal_ens_cell / len(li_metal)

# %%
"""
Make a list of equilibrium potentials.
"""

# %%
# V_eq_ens = lifepo4_ens - ...

# teacher
V_eq_ens = lifepo4_ens - fepo4_ens - li_metal_ens

# %%
"""
Use the plot command below to visualize the distribution.
"""

# %%
plt.hist(V_eq_ens, 50)
plt.grid(True)

# %%
"""
Use the [NumPy function standard deviation function](https://docs.scipy.org/doc/numpy/reference/generated/numpy.std.html#numpy-std) to obtain the standard deviation of the ensemble.
"""

# %%
# error = ...
# print(error)

# teacher
error = np.std(V_eq_ens)
print(error)

# %%
"""
The equilibrium potential for a FePO$_4$/Li battery is thus as a good estimate:
"""

# %%
print(f'{V_eq:.2f} V +- {error:.2f} V')

# %%
"""
You can get the equilibrium potential for the FePO$_4$/C battery using the intercallation energy of Li in graphite, that you calculated on Day 2. What equilibrium potential do you find? How does that compare to the cell voltage you can obtain from FePO$_4$/C batteries?
"""

# %%
# You can use this cell for FePO4/C potential calculation

# %%
"""
Make sure you are able to discuss the bullet points at the top of this notebook. You can use the cell below for calculations.
"""

# %%


# %%
"""
## Bonus
"""

# %%
"""
How does the predicted error estimate change if you consider the full reaction from Li in graphite + FePO4  to empty graphite + LiFePO4.
"""



import numpy as np
import ase.units as units
from ase.utils.filecache import MultiFileJSONCache
from gpaw.elph.electronphonon import ElectronPhononCoupling
from gpaw.mpi import world
from gpaw.typing import ArrayND


class EPC(ElectronPhononCoupling):
    """Modify ElectronPhononCoupling to fit Raman stuff better.

    Primarily, always save the supercell matrix in separate files,
    so that calculation of the elph matrix can be better parallelised.
    """

    def calculate_supercell_matrix(self, calc, name='supercell', filter=None,
                                   include_pseudo=True):
        """
        Calculate elph supercell matrix.

        This is a necessary intermediary step before calculating the electron-
        phonon matrix.

        The Raman version always uses dump=2 when calling
        the ElectronPhononCoupling routine.

        Parameters
        ----------
        calc: GPAW
            Converged ground-state calculation. Same grid as before.
        name: str
            User specified name for supercell cache. Default: 'supercell'
        filter: str
            Fourier filter atomic gradients of the effective potential. The
            specified components (``normal`` or ``umklapp``) are removed
            (default: None).
        include_pseudo: bool
            Include the contribution from the psedupotential in the atomic
            gradients. If ``False``, only the gradient of the effective
            potential is included (default: True).
        """
        self.set_lcao_calculator(calc)
        ElectronPhononCoupling.calculate_supercell_matrix(self, name, filter,
                                                          include_pseudo)

    def _bloch_matrix(self, kpt, k_c, u_l, name) -> ArrayND:
        """
        This is a special q=0 version. Need to implement general version in
        ElectronPhononCoupling.
        """
        assert len(u_l.shape) == 3

        # Defining system sizes
        nmodes = u_l.shape[0]
        nbands = kpt.C_nM.shape[0]
        nao = kpt.C_nM.shape[1]
        ndisp = 3 * len(self.indices)

        # Lattice vectors
        R_cN = self.compute_lattice_vectors()
        # Number of unit cell in supercell
        N = np.prod(self.supercell)

        # Allocate array for couplings
        g_lnn = np.zeros((nmodes, nbands, nbands), dtype=complex)

        # Mass scaled polarization vectors
        u_lx = u_l.reshape(nmodes, ndisp)

        self.supercell_cache = MultiFileJSONCache(name)
        self.basis_info = self.supercell_cache['basis']

        # Multiply phase factors
        for x in range(ndisp):
            # Allocate array
            g_MM = np.zeros((nao, nao), dtype=complex)
            g_sNNMM = self.supercell_cache[str(x)]
            assert nao == g_sNNMM.shape[-1]
            for m in range(N):
                for n in range(N):
                    phase = self._get_phase_factor(R_cN, m, n, k_c,
                                                   [0., 0., 0.])
                    # Sum contributions from different cells
                    g_MM += g_sNNMM[kpt.s, m, n, :, :] * phase

            g_nn = np.dot(kpt.C_nM.conj(), np.dot(g_MM, kpt.C_nM.T))
            g_lnn += np.einsum('i,kl->ikl', u_lx[:, x], g_nn)

        return g_lnn * units.Hartree / units.Bohr  # eV / Ang

    def get_elph_matrix(self, calc, phonon, name='supercell',
                        savetofile=True) -> ArrayND:
        """Calculate the electronphonon matrix in Bloch states.

        Always uses q=0.

        Parameters
        ----------
        calc: GPAW
            Converged ground-state calculation. NOT supercell.
        phonon: ase.phonons.Phonons
            Phonon object
        name: str
            Name of supercell cache
        savetofile: bool
            Switch for saving to gsqklnn.npy file
        """
        assert calc.wfs.bd.comm.size == 1

        # Read previous phonon calculation.
        # This only looks at gamma point phonons
        phonon.read()
        frequencies, modes = phonon.band_structure([[0., 0., 0.]], modes=True)

        g_sqklnn = np.zeros([calc.wfs.nspins, 1, calc.wfs.kd.nibzkpts,
                             frequencies.shape[1], calc.wfs.bd.nbands,
                             calc.wfs.bd.nbands], dtype=complex)

        # loop over k-points
        for kpt in calc.wfs.kpt_u:
            k_c = calc.wfs.kd.ibzk_kc[kpt.k]
            # Find el-ph matrix in the LCAO basis
            g_lnn = self._bloch_matrix(kpt, k_c, modes[0], name)
            g_sqklnn[kpt.s, 0, kpt.k] += g_lnn

        calc.wfs.kd.comm.sum(g_sqklnn)

        if world.rank == 0 and savetofile:
            np.save("gsqklnn.npy", g_sqklnn)
        return g_sqklnn



from ase.build import bulk
from gpaw import GPAW, PW
from gpaw.xc.libvdwxc import vdw_df_cx

# "Large" system:
atoms = bulk('Cu').repeat((2, 2, 2))
calc = GPAW(mode=PW(600),
            kpts=(4, 4, 4),
            xc=vdw_df_cx(mode='pfft', pfft_grid=(2, 2)),
            parallel=dict(kpt=4, augment_grids=True))
atoms.calc = calc
atoms.get_potential_energy()



from myqueue.workflow import run


def workflow():
    with run(script='Fe_gs.py', cores=40, tmax='1h'):
        with run(script='Fe_mft.py', cores=40, tmax='2h'):
            run(script='Fe_plot_magnons_vs_rc.py')
            run(script='Fe_plot_magnon_dispersion.py')
    with run(script='Co_gs.py', cores=40, tmax='1h'):
        with run(script='Co_mft.py', cores=40, tmax='6h'):
            run(script='Co_plot_magnons_vs_rc.py')
            run(script='Co_plot_magnon_dispersion.py')



# web-page: silicon_ABS.png
import numpy as np
import matplotlib.pyplot as plt

plt.figure(figsize=(7, 5))
d = np.loadtxt('si_abs.csv', delimiter=',')
plt.plot(d[:, 0], d[:, 3], '-k', label=r'$\mathrm{Re}\epsilon(\omega)$')
plt.plot(d[:, 0], d[:, 4], '-r', label=r'$\mathrm{Im}\epsilon(\omega)$')

plt.title('Dielectric function of Si')
plt.legend()
plt.xlabel('Energy (eV)', fontsize=14)
plt.ylabel(r'$\epsilon$', fontsize=18)

ax = plt.gca()


def y(e):
    x = d[:, 0]
    i = (x < e).sum()
    return d[i, 4]


# data from G.Kresse, PRB 73, 045112 (2006)
for name, e in zip("E_0 E_1 E_2 E_0' E_1'".split(),
                   [2.53, 2.71, 3.72, 3.08, 4.50]):
    arr = plt.arrow(e, y(e) + 4.0, 0, -3,
                    width=0.01, head_width=0.1, head_length=1)
    ax.add_patch(arr)
    plt.text(e, y(e) + 4, '$' + name + '$')

plt.xlim(0, 6)
plt.ylim(0, 60)
plt.savefig('silicon_ABS.png', bbox_inches='tight')
plt.show()



import numpy as np

from ase.phonons import Phonons

from gpaw import GPAW
from gpaw.elph.electronphonon import ElectronPhononCoupling

calc = GPAW("scf.gpw")

kpts = calc.get_ibz_k_points()
qpts = [[0, 0, 0], ]

# Phonon calculation, We'll read the forces from the elph.run function
# This only looks at gamma point phonons
ph = Phonons(atoms=calc.atoms, supercell=(1, 1, 1))
ph.read()
frequencies, modes = ph.band_structure(qpts, modes=True)

# Find el-ph matrix in the LCAO basis
elph = ElectronPhononCoupling(calc.atoms, calc=None, calculate_forces=False)
elph.set_lcao_calculator(calc)
elph.load_supercell_matrix()

# Find the bloch expansion coefficients
g_sqklnn = []
for s in range(calc.wfs.nspins):
    c_kn = []
    for k in range(calc.wfs.kd.nibzkpts):
        C_nM = calc.wfs.collect_array('C_nM', k, s)
        c_kn.append(C_nM)
    c_kn = np.array(c_kn)

    # And we finally find the electron-phonon coupling matrix elements!
    # elph.g_xNNMM = elph.g_xsNNMM[:, s]
    g_qklnn = elph.bloch_matrix(kpts, qpts, c_kn, u_ql=modes, spin=s)
    g_sqklnn.append(g_qklnn)
    # np.save("gsqklnn.npy", np.array(g_sqklnn))



from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter

# Read the ground-state file
td_calc = LCAOTDDFT('gs.gpw')

# Attach analysis tools
DipoleMomentWriter(td_calc, 'dm_replayed.dat')

# Replay the propagation
td_calc.replay(name='wf.ulm', update='all')



import numpy as np

from ase.optimize import BFGS
from ase import Atoms
from ase.data.vdw import vdw_radii

from gpaw import restart
from gpaw.solvation.sjm import SJM, SJMPower12Potential
from gpaw import FermiDirac
from gpaw.solvation import (
    EffectivePotentialCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction
)


def atomic_radii(atoms):
    return [vdw_radii[n] for n in atoms.numbers]


# Solvent parameters
u0 = 0.180  # eV
epsinf = 78.36  # dielectric constant of water at 298 K
gamma = 0.00114843767916  # 18.4*1e-3 * Pascal*m
T = 298.15    # K

atoms = Atoms(symbols='Au6OH2OH2OH2OH2',
              pbc=[True, True, False],
              cell=[8.867119036079306, 5.119433562416841, 15.0])
atoms.set_positions(
    np.array([[0.56164375, 0.39256423, 5.4220237],
              [3.51680174, 0.39253274, 5.42219168],
              [6.4732849, 0.39549906, 5.42907972],
              [2.03914701, 2.95449556, 5.42846579],
              [4.99533678, 2.95180933, 5.42125025],
              [7.9504729, 2.95176062, 5.42333033],
              [6.58498136, 0.64530367, 9.09028872],
              [7.06181382, -0.19870328, 9.30204832],
              [6.58872823, 0.67663023, 8.10927953],
              [7.99675219, 3.3136834, 9.76193639],
              [7.58790792, 2.45509461, 9.51090101],
              [8.93840968, 3.2284547, 9.49325933],
              [3.56839946, 0.71728602, 9.69844062],
              [3.16824247, -0.14379405, 9.43936165],
              [4.51456907, 0.64768502, 9.44106286],
              [2.14806966, 3.19083417, 9.07185988],
              [2.62667731, 2.34458289, 9.26969167],
              [2.10900492, 3.20816497, 8.09062402]]))

sj = {'target_potential': 4.5,
      'excess_electrons': 0.124,
      'jelliumregion': {'top': 14.},
      'tol': 0.005}

calc = SJM(sj=sj,
           gpts=(48, 32, 88),
           kpts=(2, 2, 1),
           xc='PBE',
           occupations=FermiDirac(0.1),
           cavity=EffectivePotentialCavity(
               effective_potential=SJMPower12Potential(atomic_radii, u0,
                                                       H2O_layer=True),
               temperature=T,
               surface_calculator=GradientSurface()),
           dielectric=LinearDielectric(epsinf=epsinf),
           interactions=[SurfaceInteraction(surface_tension=gamma)],
           txt='sjm.txt',
           convergence={'energy': 0.000005})

atoms.calc = calc
E = []

for pot in [4.5, None, 4.3, 4.5]:
    if pot is None:
        calc.set(sj={'excess_electrons': 0.2, 'target_potential': None})
    else:
        calc.set(sj={'target_potential': pot})
    E.append(atoms.get_potential_energy())

    if pot is None:
        assert abs(calc.wfs.nvalence - calc.setups.nvalence - 0.2) < 1e-4
    else:
        assert abs(calc.get_electrode_potential() - pot) < 0.005

assert abs(E[0] - E[-1]) < 1e-2

calc.write('sjm.gpw')

atoms, calc = restart('sjm.gpw', Class=SJM)
calc.set(sj={'tol': 0.002})
atoms.get_potential_energy()
assert abs(calc.get_electrode_potential() - 4.5) < 0.002

calc.set(sj={'jelliumregion': {'top': 13}, 'tol': 0.01})
atoms.get_potential_energy()
assert abs(calc.get_electrode_potential() - 4.5) < 0.01

calc.set(sj={'jelliumregion': {'thickness': 2}})
atoms.get_potential_energy()
assert abs(calc.get_electrode_potential() - 4.5) < 0.01

qn = BFGS(atoms, maxstep=0.05)
qn.run(fmax=0.05, steps=2)
assert abs(calc.get_electrode_potential() - 4.5) < 0.01



define_macros += [('GPAW_NO_UNDERSCORE_BLAS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_LAPACK', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_BLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_SCALAPACK', '1')]
define_macros += [('GPAW_ASYNC', 1)]
define_macros += [('GPAW_MPI2', 1)]
# define_macros += [('GPAW_MR3',1)] # requires developmental ScaLAPACK
# uncomment two lines below for FLOP rate measurement
# define_macros += [('GPAW_HPM',1)]
define_macros += [('GPAW_PERFORMANCE_REPORT',1)]
# define_macros += [('GPAW_MPI_DEBUG', 1)] # debugging
# define_macros += [('GPAW_OMP',1)] # not really working

scalapack = True
hdf5 = True

# If you are using threading, you probably
# need to change the following library:
# xlomp_ser -> xlsmp
#
# DO NOT INTERCHANGE THE ORDER OF LAPACK
# & ESSL, LAPACK SHOULD BE LINKED FIRST.
#
# Goto appears to be much faster for general
# DGEMM operations, particularly those with:
# alpha != 1.0 and beta != 0.0
#
# Goto is hand-tuned assembly, it will most
# likely always be faster than ESSL-4.x.
# NAR: Goto appears to cause core dumps for
# some problems, use at your own risk.
# Disabling the stackground seems to make
# the problem go away, but this is not 
# recommended.
# --env BG_STACKGUARDENABLE=0
#multi threaded
libraries = [
           'scalapack',
           'lapack',
           'esslsmpbg',
           'xlf90_r',
           'xlopt',
           'xl',
           'xlfmath',
           'xlsmp',
            ]

#single threaded
# libraries = [
#            'scalapack',
#            'lapack',
#            'esslbg',
#            'xlf90_r',
#            'xlopt',
#            'xl',
#            'xlfmath',
#            'xlomp_ser',
#             ]

import os
ibmcmp_base = os.environ['ibmcmp_base']
python_base = '/soft/apps/python/scalable-python-2.6.7-cnk-gcc'

library_dirs = [
           '/soft/libraries/alcf/current/xl/LAPACK/lib',
           '/soft/libraries/alcf/current/xl/SCALAPACK/lib',
           '/soft/libraries/essl/5.1.1-0/lib64',
           f'{ibmcmp_base}/xlf/bg/14.1/bglib64',
           f'{ibmcmp_base}/xlsmp/bg/3.1/bglib64',
# plain vanilla Python
#           '/bgsys/tools/Python-2.6/lib64',
# scalable Python 2.6.7
           f'{python_base}',
           '/soft/libraries/unsupported/hdf5-1.8.8/lib/',
           ]

# plain vanilla Python
# include_dirs += [
#    '/soft/apps/python/python-2.6.6-cnk-gcc/bgsys/tools/Python-2.6/lib64/python2.6/site-packages/numpy/core/include'
#    ]

# scalable Python 2.6.7
include_dirs += [
    f'{python_base}/lib/python2.6/site-packages/numpy/core/include',
    '/soft/libraries/unsupported/hdf5-1.8.8/include/'
    ]

mpi_libraries = [
#    'mpihpm_smp',
    'hdf5',
    'mpich',
    'opa',
    'mpl',
    'pami',
    'SPI',
    'SPI_cnk',
    'stdc++',
#    'bgpm',
    ]

mpi_library_dirs = [
    '/bgsys/drivers/ppcfloor/comm/xl.legacy/lib',
    '/bgsys/drivers/ppcfloor/comm/sys/lib',
    '/bgsys/drivers/ppcfloor/spi/lib',
    '/soft/perftools/hpctw',
    '/soft/perftools/bgpm/lib',
    ]

extra_link_args = ['-Wl,-export-dynamic']
compiler = "./bgq_xlc.py"
mpicompiler = "./bgq_xlc.py"
mpilinker = "./bgq_xlc_linker.py"



import gpaw.solvation as solv

# ...

# convenient way to use HW14 water parameters:
calc = solv.SolvationGPAW(
    xc='PBE', h=0.2,  # non-solvent DFT parameters
    **solv.get_HW14_water_kwargs()
)

# ...



from ase.build import bulk
from gpaw import GPAW, FermiDirac, PW

a = bulk('Fe', 'bcc')
a.set_initial_magnetic_moments([-2.0])
a.set_cell([[1.434996, 1.434996, 1.434996],
            [-1.434996, 1.434996, 1.434996],
            [-1.434996, -1.434996, 1.434996]])
calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(width=0.01),
            kpts=(12, 12, 12),
            txt='Fe_scf.txt')
a.calc = calc
a.get_potential_energy()

calc.fixed_density(
    kpts={'size': (8, 8, 8), 'gamma': True},
    symmetry='off',
    nbands=40,
    convergence={'bands': 30},
    txt='Fe_nscf.txt').write('Fe.gpw', mode='all')



import pytest
from gpaw.doctools.codegraph import main


@pytest.mark.ci
@pytest.mark.serial
def test_code_graph(monkeypatch):
    pytest.importorskip('graphviz')
    monkeypatch.setattr('graphviz.Digraph.render',
                        lambda self, name, format: None)
    main()



from myqueue.workflow import run


def workflow():
    with run(script='top.py', cores=8, tmax='15m'):
        run(script='pdos.py')
    with run(script='lcaodos_gs.py', cores=8, tmax='15m'):
        run(script='lcaodos_plt.py')



"""Calculation utilizing RSF with optimized gamma."""
from ase import Atoms
from gpaw import GPAW, setup_paths
from gpaw.poisson import PoissonSolver
from gpaw.eigensolvers import RMMDIIS
from gpaw.occupations import FermiDirac
from gpaw.test import equal, gen

# IP for CO using LCY-PBE with gamma=0.81 after
# dx.doi.org/10.1021/acs.jctc.8b00238
IP = 14.31

if setup_paths[0] != '.':
    setup_paths.insert(0, '.')

for atom in ['C', 'O']:
    gen(atom, xcname='PBE', scalarrel=True, exx=True,
        yukawa_gamma=0.81)

h = 0.30
co = Atoms('CO', positions=[(0, 0, 0), (0, 0, 1.15)])
co.minimal_box(5)

# c = {'energy': 0.005, 'eigenstates': 1e-4}  # Usable values
c = {'energy': 0.1, 'eigenstates': 3, 'density': 3}  # Values for test

calc = GPAW(txt='CO.txt', xc='LCY-PBE:omega=0.81', convergence=c,
            eigensolver=RMMDIIS(), h=h,
            poissonsolver=PoissonSolver(use_charge_center=True),
            occupations=FermiDirac(width=0.0), spinpol=False)
co.calc = calc
co.get_potential_energy()
(eps_homo, eps_lumo) = calc.get_homo_lumo()
equal(eps_homo, -IP, 0.15)
